package examples.hospital;

import ga.*;
import helperclasses.*;
import ttclasses.*;
import java.io.*;
import java.util.*;


/*****
 * This class manages the solution of TTProblems of the hospital type.
 * There are four example problems of this type: chirurgie, geriatrie, innere and innere2.
 *****/
public class Hospital extends TTProblem {
  String dataFile;

  /** Creates new <tt>Hospital</tt>, initializes local data with what can be found in <tt>dataDir</tt. */
  public Hospital(String xmlFile) {
    random = GlobalObjects.random;
    dataFile = xmlFile;

    constraints.add(new DiffHourMonthConstraint());
    constraints.add(new OnlyOneDayFreeConstraint());
    constraints.add(new OnlyOneWorkingDayConstraint());
    constraints.add(new OutOfMinMaxConstraint());
    constraints.add(new ClashConstraint());
    constraints.add(new LittleFreeTimeConstraint());
    
    constraints.add(new ContinuationBonus());

    initialize(xmlFile);
  } // end of Hospital(String)

  public final TTProblem createNewInstance() {
    return new Hospital(dataFile);
  } // end of createNewInstance()

  /**
   * Since all events and resources can be created dynamically while processing
   * the XML data file belonging to the GTL description of the problem, <tt>TTProblem</tt>
   * must be able to dynamically instantiate objects of certain types.
   */
  public Event createEvent(String type, String name) {
    Event result = null;
    if (type.equals("Shift"))
      result = new Shift(name);
    if (result != null)
      eventList.add(result);
    return result;
  } // end of createEvent(String, String)

  /**
   * Since all events and resources can be created dynamically while processing
   * the XML data file belonging to the GTL description of the problem, <tt>TTProblem</tt>
   * must be able to dynamically instantiate objects of certain types.
   */
  public Resource createResource(String type, String name) {
    Resource result = null;
    List rt = getResourceType(type);
    if (rt == null)
      rt = addResourceType(type);
    if (type.equals("Staff"))
      result = new Staff(name);
    else
      result = super.createResource(type, name);
    if (result != null)
      rt.add(result);
    return result;
  } // end of createResource(String, String)

  public static void main(String args[]) {
    System.out.println("Which timetable shall be created?");
    System.out.println("(1) Geriatrie   (2) Innere");
    System.out.print("> ");
    try {
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      String selection = br.readLine();
      switch (Integer.parseInt(selection)) {
        case 1: selection = "../databases/xml/geriatrie_februar.xml"; break;
        case 2: selection = "../databases/xml/innere_februar.xml"; break;
        default: selection = null;
      }
      Hospital h = new Hospital(selection);
      h.solve(new GeneticAlgorithm());
    }
    catch (Exception e) {
      System.out.println("Something went wrong.");
    }
  } // end of main(String[])
  
} // end of class Hospital

/** Constraint descriptions can be found in the paper. **/
class OutOfMinMaxConstraint extends HardConstraint {

  /** A boring constructor. **/  
  public OutOfMinMaxConstraint() {
    name = "OutOfMinMaxConstraint";
  }
  
  public ConstraintViolation compute(PlanningClass owner) {
    double akku = 0;
    Event e = (Event)owner;
    List list = new ArrayList();
    for (int i = e.resources.size() - 1; i >= 0; i--) {
      ResourceList resList = (ResourceList)e.resources.get(i);
      if (!resList.getName().equals("TimeSlots")) {   // consider all reslists except the timeslots
        int assigned = resList.assignedCount();
        if (assigned > resList.getMax()) {
          akku += 200 * (assigned - resList.getMax());
          list.add(resList);
        }
        else if (assigned < resList.getMin()) {
          list.add(resList);
          akku += 200 * (resList.getMin() - assigned);
        }
      }
    }
    if (akku > 0) {
      e.penaltyPoints += akku;
      return new ConstraintViolation(this, owner, "MinMaxViolation at " + owner, akku, list);
    }
    else
      return null;
  } // end of compute(PlanningClass)
 
  /**
   * Violations of this constraint will be repaired by randomly
   * assigning / freeing resources.
   **/ 
  public boolean repair(ConstraintViolation violation) {
/*
    List list = (List)violation.information;
    for (int i = list.size() - 1; i >= 0; i--) {
      ResourceList rl = (ResourceList)list.get(i);
      while (rl.assignedCount() > rl.getMax())
        rl.freeRandom();
      while (rl.assignedCount() < rl.getMin())
        rl.assignRandom();
    }
*/
    return true;
  } // end of repair(ConstraintViolation)
  
} // end of class OutOfMinMaxConstraint


class DiffHourMonthConstraint extends SoftConstraint {

  public DiffHourMonthConstraint() {
    name = "DiffHourMonthConstraint";
  }
  
  public ConstraintViolation compute(PlanningClass owner) {
    Staff res = (Staff)owner;
    double akku = -4 * res.workLoad;
    List events = res.getEvents();
    // sum actual work load over all events that the person has been assigned to
    for (int i = events.size() - 1; i >= 0; i--) {
      Event e = (Event)events.get(i);
      ResourceList resList = e.getResourceList("TimeSlots");
      List assignment = resList.getAssignmentList();
      for (int j = assignment.size() - 1; j >= 0; j--) {
        TimeInterval ti = (TimeInterval)assignment.get(j);
        akku += ((double)ti.getDuration()) / 60;
      }
    }
    if (akku > 0)
      return new ConstraintViolation(this, owner, "DiffHourMonthViolation at " + res.name + " (+)", akku * 12);
    else if (akku < 0)
      return new ConstraintViolation(this, owner, "DiffHourMonthViolation at " + res.name + " (-)", -akku * 12);
    else
      return null;
  } // end of compute(PlanningClass)
 
  /** No repair method implemented. **/ 
  public boolean repair(ConstraintViolation violation) {
    return false;
  }
  
} // end of class DiffHourMonthConstraint


class OnlyOneDayFreeConstraint extends SoftConstraint {

  public OnlyOneDayFreeConstraint() {
    name = "OnlyOneDayFreeConstraint";
  }

  public ConstraintViolation compute(PlanningClass owner) {
    double akku = 0;
    int j, index, cnt;
    Staff staff = (Staff)owner;
    boolean days[] = new boolean[32];
    List timeSlots = staff.getTimeSlots();
    // first, assume the resource is never occupied
    for (j = 0; j < 32; j++)
      days[j] = false;
    for (j = timeSlots.size() - 1; j >= 0; j--) {
      // now, set day states according to assignment list
      TimeInterval ti = (TimeInterval)timeSlots.get(j);
      days[ti.from.day] = true;
      days[ti.until.day] = true;
    }
    index = 1;
    while ((index < 32) && (days[index] == days[1]))
      index++;
    cnt = 1;
    while (index < 32) {
      if (days[index] == days[index - 1])  // days of same type
        cnt++;
      else {  // working->free  or  free->working
        if (days[index]) {  // working day
          if ((cnt == 1) && (index > 2))  // only one free day => Penalty
            akku += 150;
        }
        cnt = 1;
      }
      index++;
    }
    if (akku > 0)
      return new ConstraintViolation(this, owner, "OnlyOneDayFree at " + staff.name, akku);
    else
      return null;
  } // end of compute(PlanningClass)

  /** No repair method implemented. **/
  public boolean repair(ConstraintViolation violation) {
    return false;
  }
  
} // end of class OnlyOneDayFreeConstraint


/*****
 * If a there is only one working day between two days of holiday in a person's timetable,
 * that situation will be punished by this constraint.
 *****/
class OnlyOneWorkingDayConstraint extends SoftConstraint {

  public OnlyOneWorkingDayConstraint() {
    name = "OnlyOneWorkingDayConstraint";
  }

  public ConstraintViolation compute(PlanningClass owner) {
    double akku = 0;
    int j, index, cnt;
    Staff staff = (Staff)owner;
    boolean days[] = new boolean[32];
    List timeSlots = staff.getTimeSlots();
    // first, assume the resource is never occupied
    for (j = 0; j < 32; j++)
      days[j] = false;
    for (j = timeSlots.size() - 1; j >= 0; j--) {
      // now, set day states according to assignment list
      TimeInterval ti = (TimeInterval)timeSlots.get(j);
      days[ti.from.day] = true;
      days[ti.until.day] = true;
    }
    index = 1;
    while ((index < 32) && (days[index] == days[1]))
      index++;
    cnt = 1;
    while (index < 32) {
      if (days[index] == days[index - 1])  // days of same type
        cnt++;
      else {  // working->free  or  free->working
        if (!days[index]) {  // working day
          if ((cnt == 1) && (index > 2))  // only one free day => Penalty
            akku += 150;
        }
        cnt = 1;
      }
      index++;
    }
    if (akku > 0)
      return new ConstraintViolation(this, owner, "OnlyOneWorkingDay at " + staff.name, akku);
    else
      return null;
  } // end of compute(PlanningClass)
  
  /** No repair method implemented. **/
  public boolean repair(ConstraintViolation violation) {
    return false;
  }
  
} // end of class OnlyOneWorkingDayConstraint


/*****
 * This constraint checks if there is enough distance between two working
 * periods of a given person (min 10 hours).
 *****/
class LittleFreeTimeConstraint extends HardConstraint {
  private Random random;

  class TemporalEventComparator implements Comparator {

    public final int compare(java.lang.Object o1, java.lang.Object o2) {
      try {
        List list1 = ((Event)o1).getAssignedTimeSlots();
        List list2 = ((Event)o2).getAssignedTimeSlots();
        if (list1.size() == 0) {
          if (list2.size() == 0)
            return 0;
          else
            return -1;
        }
        else {
          TimeInterval ti1 = (TimeInterval)list1.get(0);
          TimeInterval ti2 = (TimeInterval)list2.get(0);
          return (ti1.from.compareTo(ti2.from));
        }
      } catch (Exception e) {
        return 0;
      }
    }

    public final boolean equals(java.lang.Object obj) {
      try {
        TemporalEventComparator tec = (TemporalEventComparator)obj;
        return true;
      } catch (Exception e) {
        return false;
      }
    }

  } // end of inner class TemporalEventComparator

  public LittleFreeTimeConstraint() {
    name = "LittleFreeTimeConstraint";
    random = GlobalObjects.random;
  }

  /** Returns the temporal distance between two events in hours. **/
  private int distance(Event e1, Event e2) {
    List list1 = e1.getAssignedTimeSlots();
    List list2 = e2.getAssignedTimeSlots();
    if ((list1.size() == 0) || (list2.size() == 0))
      return 999;
    TimeInterval ti1 = (TimeInterval)(list1.get(0));
    TimeInterval ti2 = (TimeInterval)(list2.get(0));
    return ti1.distance(ti2) / 60;
  }

  public ConstraintViolation compute(PlanningClass owner) {
    double akku = 0;
    Staff staff = (Staff)owner;
    Object events[] = staff.getEvents().toArray();
    Arrays.sort(events, new TemporalEventComparator());
    ArrayList info = new ArrayList();
    int length = events.length;
    for (int i = 0; i < length - 1; i++)
      if (distance((Event)events[i], (Event)events[i + 1]) < 10) {
        akku += 180;
        ((Event)events[i]).penaltyPoints += 90;
        ((Event)events[i + 1]).penaltyPoints += 90;
        info.add(events[i]);
        info.add(events[i + 1]);
      }
    if (akku > 0)
      return new ConstraintViolation(this, owner, "LittleFreeTime at " + staff.name, akku, info);
    else
      return null;
  } // end of compute(PlanningClass)

  /**
   * The repair method works by randomly freeing the resource at
   * one of the events causing the trouble.
   **/
  public boolean repair(ConstraintViolation violation) {
    ArrayList list = (ArrayList)violation.information;
    Resource res = (Resource)violation.owner;
    int index = random.nextInt(list.size());
    Event ev = (Event)list.get(index);
    ev.freeResource(res);
    return true;
  } // end of repair(ConstraintViolation)

} // end of class LittleFreeTimeConstraint


/*****
 * This is an inverse constraint that donates some bonus points if a person
 * is assigned to the same shift type on succeeding days.
 *****/
class ContinuationBonus extends SoftConstraint {

  public ContinuationBonus() {
    name = "ContinuationBonus";
  }

  public ConstraintViolation compute(PlanningClass owner) {
    Staff staff = (Staff)owner;
    int i, akku;
    int shiftTypes[] = new int[32];
    for (i = 0; i < 32; i++)
      shiftTypes[i] = -1;
    List events = staff.getEvents();
    for (i = events.size() - 1; i >= 0; i--) {
      Shift shift = (Shift)events.get(i);
      List timeslots = shift.getAssignedTimeSlots();
      TimeInterval ti = (TimeInterval)timeslots.get(0);
      shiftTypes[ti.from.day] = shift.shiftType;
    }
    akku = 0;
    for (i = 0; i < 31; i++)
      if (shiftTypes[i] == shiftTypes[i + 1])
        akku++;
    if (akku == 0)
      return null;
    else
      return new ConstraintViolation(this, owner, "Continuation Bonus for " + staff.name, - akku * 20);
  }

  /** We do not want to repair "violations" of this constraint. **/
  public boolean repair(ConstraintViolation violation) {
    return false;
  }

} // end of ContinuationBonus



