package examples.maedchen;

import ga.*;
import helperclasses.*;
import ttclasses.*;
import java.io.*;
import java.util.*;


/*****
 * This class manages the solution of TTProblems of the type
 * "Maedchen+Technik-Praktikum". It can operate on the databases
 * of different years.
 *****/
public class MaedchenUndTechnik extends TTProblem {
  private String kurse, kursplan, teilnehmer;
  private List projectTypes;

  /**
   * First constructor: Initializes problem from XML data input.
   **/
  public MaedchenUndTechnik(String dataFile) {
    // initialize constraints and store in constraint list
    constraints.add(new MinMaxConstraint());
    constraints.add(new ClashConstraint());
    constraints.add(new DoubleSignatureConstraint());
    constraints.add(new FriendsConstraint());
    constraints.add(new HardProjectConstraint());
    constraints.add(new SoftChoiceConstraint());
    constraints.add(new SoftProjectConstraint());
    
    initialize(dataFile);
  }

  /**
   * Second constructor: Works directly on the "Maedchen und Technik" database files.
   * Creates new <tt>MaedchenUndTechnik</tt>, initializes local data.
   **/
  public MaedchenUndTechnik(String kurse, String kursplan, String teilnehmer) {
    random = GlobalObjects.random;
    TimeInterval slots[];
    List list;
    
    this.kurse = kurse;
    this.kursplan = kursplan;
    this.teilnehmer = teilnehmer;
    
    /*************************/
    /* create all time slots */
    /*************************/
    
    slots = new TimeInterval[10];
    slots[0] = new TimeInterval(new Time(1, 9, 0), new Time(1, 12, 0), "Montag Vormittag");
    slots[1] = new TimeInterval(new Time(1, 13, 0), new Time(1, 16, 0), "Montag Nachmittag");
    slots[2] = new TimeInterval(new Time(2, 9, 0), new Time(2, 12, 0), "Dienstag Vormittag");
    slots[3] = new TimeInterval(new Time(2, 13, 0), new Time(2, 16, 0), "Dienstag Nachmittag");
    slots[4] = new TimeInterval(new Time(3, 9, 0), new Time(3, 12, 0), "Mittwoch Vormittag");
    slots[5] = new TimeInterval(new Time(3, 13, 0), new Time(3, 16, 0), "Mittwoch Nachmittag");
    slots[6] = new TimeInterval(new Time(4, 9, 0), new Time(4, 12, 0), "Donnerstag Vormittag");
    slots[7] = new TimeInterval(new Time(4, 13, 0), new Time(4, 16, 0), "Donnerstag Nachmittag");
    slots[8] = new TimeInterval(new Time(5, 9, 0), new Time(5, 12, 0), "Freitag Vormittag");
    slots[9] = new TimeInterval(new Time(5, 13, 0), new Time(5, 16, 0), "Freitag Nachmittag");

    // now, we set the typical weekly-timetable design so that
    // the timetable can be written to HTML
    ttXlayout = 5; ttYlayout = 2; ttZlayout = 1;
    slots[0].setLayout(1, 1, 1);
    slots[1].setLayout(1, 2, 1);
    slots[2].setLayout(2, 1, 1);
    slots[3].setLayout(2, 2, 1);
    slots[4].setLayout(3, 1, 1);
    slots[5].setLayout(3, 2, 1);
    slots[6].setLayout(4, 1, 1);
    slots[7].setLayout(4, 2, 1);
    slots[8].setLayout(5, 1, 1);
    slots[9].setLayout(5, 2, 1);

    list = addResourceType("TimeSlot");
    for (int i = 0; i < 10; i++)
      list.add(slots[i]);

    BufferedReader br;
    projectTypes = new ArrayList();

    /* Read data from file "Kurse.txt" in order to create event classes (project types). */
    
    br = createReader(kurse);
    if (br != null) {  // begin file processing
      String line, name;
      int j, slotnumber;
      while ((line = readLine(br)) != null) {
        // every line in file "Kurse.txt" has the following format: "projectname min/max*timeslots".
        line = line.trim();
        j = line.indexOf(' ');
        if (j > 0) {
          name = line.substring(0, j);
          line = line.substring(j + 1);
          j = line.indexOf('*');
          if (j > 0) {
            try {
              slotnumber = Integer.parseInt(line.substring(j + 1));
            }
            catch (Exception e1) {
              System.err.println("Error parsing file \"" + kurse + "\".");
              slotnumber = 1;
            }
            line = line.substring(0, j);
            j = line.indexOf('/');
            if (j > 0) {
              try {
                projectTypes.add(new ProjectType(name, Integer.parseInt(line.substring(0, j)),
                                 Integer.parseInt(line.substring(j + 1)), slotnumber));
              }
              catch (Exception e2) {
                System.err.println("Error parsing file \"" + kurse + "\".");
              }
            }
          }
        }
      }
    } // end file processing "Kurse.txt"

    br = createReader(kursplan);
    if (br != null) {  // begin processing file "Kursplan.txt"
      String currentDay = "Sonntag";
      int currentTimeslot = 0;
      String line, name, type;
      int j, slotnumber;
      while ((line = readLine(br)) != null) {
        line = line.trim();
        j = line.indexOf(' ');
        if (j > 0) {  // ordinary line containing project a description
          type = line.substring(0, j);
          line = line.substring(j + 1);
          if (line.length() == 2) {  // project lines are of the format "type VN", "type V_" or "type _N"
            ProjectType pt = getProjectType(type);
            if (pt != null) {
              if ((pt.slotnumber == 2) && (line.equals("VN"))) {
                name = type + " (" + currentDay + ")";
                list = new ArrayList(); list.add(slots[currentTimeslot]); list.add(slots[currentTimeslot + 1]);
              }
              else if ((pt.slotnumber == 1) && (line.equals("V_"))) {
                name = type + " (" + currentDay + " Vormittag)";
                list = new ArrayList(); list.add(slots[currentTimeslot]);
              }
              else if ((pt.slotnumber == 1) && (line.equals("_N"))) {
                name = type + " (" + currentDay + " Nachmittag)";
                list = new ArrayList(); list.add(slots[currentTimeslot]);
              }
              else if ((pt.slotnumber == 2) && (line.equals("_N"))) {  // this is difficult: how to deal with events
                                                                     // that are spread over two different days?
                if (pt.incompleteTimeslots == null) {  // store timeslot for later processing
                  pt.incompleteTimeslots = slots[currentTimeslot + 1];
                  name = null;
                }
                else {
                  name = type + " (" + currentDay + ")";
                  list = new ArrayList(); list.add(slots[currentTimeslot + 1]); list.add(pt.incompleteTimeslots);
                  pt.incompleteTimeslots = null;
                }
              }
              else if ((pt.slotnumber == 2) && (line.equals("V_"))) {
                if (pt.incompleteTimeslots == null) {  // store timeslot for later processing
                  pt.incompleteTimeslots = slots[currentTimeslot];
                  name = null;
                }
                else {
                  name = type + " (" + currentDay + ")";
                  list = new ArrayList(); list.add(slots[currentTimeslot]); list.add(pt.incompleteTimeslots);
                  pt.incompleteTimeslots = null;
                }
              }
              else {
                if ((pt.slotnumber == 1) && (line.equals("VN"))) {
                  Project p;
                  p = new Project(type + " (" + currentDay + " Vormittag)", type);
                  p.resources.add(new ResourceList(p, "TimeSlots", slots[currentTimeslot]));
                  eventList.add(p);
                  p = new Project(type + " (" + currentDay + " Nachmittag)", type);
                  p.resources.add(new ResourceList(p, "TimeSlots", slots[currentTimeslot + 1]));
                  eventList.add(p);
                }
                else
                  System.err.println("Error: Inconsistent (illegible) definition data. " + line + " != " + pt.slotnumber +
                                     " (" + type + "/" + currentDay + ").");
                name = null;
              }
              if (name != null) {
                Project p = new Project(name, type);
                p.resources.add(new ResourceList(p, "TimeSlots", list));
                eventList.add(p);
              }
            }
            else {  // there is no project type definition for this project => ERROR!
              System.err.println("Error: Inconsistent definition data.");
            }
          }
        }
        else { // this is a special line, e.g indicating a new day
          if (line.equals("Montag")) { currentDay = "Montag"; currentTimeslot = 0; }
          else if (line.equals("Dienstag")) { currentDay = "Dienstag"; currentTimeslot = 2; }
          else if (line.equals("Mittwoch")) { currentDay = "Mittwoch"; currentTimeslot = 4; }
          else if (line.equals("Donnerstag")) { currentDay = "Donnerstag"; currentTimeslot = 6; }
          else if (line.equals("Freitag")) { currentDay = "Freitag"; currentTimeslot = 8; }
        }
      }
    } // end of file processing "Kursplan.txt"
    
    list = addResourceType("Girls");

    br = createReader(teilnehmer);
    if (br != null) { // begin file processing "Teilnehmer.txt"
      int j;
      String line, name, firstChoice, secondChoice, friends;
      List tempList;
      while ((line = readLine(br)) != null) {  // lines are of the format "Name*FirstChoices*SecondChoices*Friends*"
        if ((j = line.indexOf('*')) <= 0) continue;
        name = line.substring(0, j);
        line = line.substring(j + 1);
        if ((j = line.indexOf('*')) <= 0) continue;
        firstChoice = line.substring(0, j);
        line = line.substring(j + 1);
        if ((j = line.indexOf('*')) < 0) continue;
        secondChoice = line.substring(0, j);
        line = line.substring(j + 1);
        if ((j = line.indexOf('*')) < 0) continue;
        friends = line.substring(0, j);
        line = line.substring(j + 1);
        
        Girl g = new Girl(name);
        tempList = stringToList(firstChoice, ',');
        for (j = 0; j < tempList.size(); j++)
          g.addFeature((String)tempList.get(j));
        g.firstChoice = tempList;
        tempList = stringToList(secondChoice, ',');
        for (j = 0; j < tempList.size(); j++)
          g.addFeature((String)tempList.get(j));
        g.friends = stringToList(friends, ';');
        
        list.add(g);
      }
    } // end file processing "Teilnehmer.txt"
    
    // now, assign girls to projects

    for (int i = 0; i < projectTypes.size(); i++) {
      ProjectType pt = (ProjectType)projectTypes.get(i);
      list = getResourcesWithFeature("Girls", pt.name);
      pt.maedchen = list;
    }

    for (int i = 0; i < eventList.size(); i++) {
      Project p = (Project)eventList.get(i);
      ProjectType pt = getProjectType(p.signature);
      if (pt != null)
        p.resources.add(new ResourceList(p, "Girls", pt.min, pt.max, pt.maedchen));
    }

    // finally, set up all necessary constraints

    Constraint c1, c2, c3;
    int i;
    
    c1 = new MinMaxConstraint();
    constraints.add(c1);
    for (i = eventList.size() - 1; i >= 0; i--) {
      ((Event)eventList.get(i)).addConstraint(c1);
    }

    List girls = getResourceType("Girls");
    
    c1 = new HardProjectConstraint();  // too many projects
    c2 = new SoftProjectConstraint();  // too few projects
    c3 = new ClashConstraint();  // clash!
    constraints.add(c1);
    constraints.add(c2);
    constraints.add(c3);
    for (i = girls.size() - 1; i >= 0; i--) {
      ((Girl)girls.get(i)).addConstraint(c1);
      ((Girl)girls.get(i)).addConstraint(c2);
      ((Girl)girls.get(i)).addConstraint(c3);
    }
    c1 = new DoubleSignatureConstraint();  // two (or more) projects of the same type
    c2 = new SoftChoiceConstraint();  // second choice
    c3 = new FriendsConstraint();
    constraints.add(c1);
    constraints.add(c2);
    constraints.add(c3);
    for (i = girls.size() - 1; i >= 0; i--) {
      ((Girl)girls.get(i)).addConstraint(c1);
      ((Girl)girls.get(i)).addConstraint(c2);
      ((Girl)girls.get(i)).addConstraint(c3);
    }

    projectTypes = null;
  } // end of MaedchenUndTechnik(String, String, String)

  /** See class <tt>TTProblem</tt> for details. @see TTProblem **/
  public Resource createResource(String type, String name) {
    Resource result = null;
    List rt = getResourceType(type);
    if (rt == null)
      rt = addResourceType(type);
    if (type.equals("Girl"))
      result = new Girl(name);
    else
      result = super.createResource(type, name);
    if (result != null)
      rt.add(result);
    return result;
  } // end of createResource(String, String)

  /** See class <tt>TTProblem</tt> for details. @see TTProblem **/
  public Event createEvent(String type, String name) {
    Event result = null;
    if (type.equals("Project"))
      result = new Project(name);
    if (result != null)
      eventList.add(result);
    return result;
  } // end of createEvent(String, String)

  /** <tt>getProjectType</tt> returns the project type with the given name. **/
  private ProjectType getProjectType(String name) {
    int k = projectTypes.size();
    for (int j = 0; j < k; j++) {
      ProjectType pt = (ProjectType)projectTypes.get(j);
      if (pt.name.equals(name))
        return pt;
    }
    return null;
  } // end of getProjectType(String)

  private List stringToList(String s, char separator) {
    List result = new ArrayList();
    int j;
    while ((j = s.indexOf(separator)) >= 0) {
      result.add(s.substring(0, j));
      s = s.substring(j + 1);
    }
    if (s.length() > 0) result.add(s);
    return result;
  } // end of stringToList(String, char)
  
  /**
   * This method creates a <tt>BufferedReader</tt> that reads data from the file that
   * is specified by <tt>filename</tt>.
   * @return <tt>null</tt> on error, the newly created <tt>BufferedReader</tt> else.
   **/
  private BufferedReader createReader(String filename) {
    try {
      FileReader fr = new FileReader(filename);
      BufferedReader br = new BufferedReader(fr);
      return br;
    }
    catch (IOException ioe) {
      System.err.println(ioe.toString());
      return null;
    }
  } // end of createReader(String)

  /** This method reads a line from the <tt>BufferedReader</tt> given and returns it to the caller. **/
  private String readLine(BufferedReader br) {
    try {
      return br.readLine();
    }
    catch (IOException ioe) {
      System.err.println(ioe.toString());
      return null;
    }
  } // end of readLine(BufferedReader)

  public final TTProblem createNewInstance() {
    return new MaedchenUndTechnik(kurse, kursplan, teilnehmer);
  } // end of createNewInstance()
  
  public static void main(String args[]) {
    System.out.println("Which year's timetable shall be created?");
    System.out.println("(1) 2000   (2) 2001   (3) 2002");
    System.out.print("> ");
    try {
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      String selection = br.readLine();
      switch (Integer.parseInt(selection)) {
        case 1: selection = "../databases/maedchen/2000/"; break;
        case 2: selection = "../databases/maedchen/2001/"; break;
        case 3: selection = "../databases/maedchen/2002/"; break;
        default: selection = null;
      }
      MaedchenUndTechnik mut = new MaedchenUndTechnik(selection + "Kurse.txt",
                                   selection + "Kursplan.txt", selection + "Teilnehmer.txt");
      mut.solve(new GeneticAlgorithm());
    }
    catch (Exception e) {
      System.out.println("Something went wrong.");
    }
  }

  
} // end of class MaedchenUndTechnik





/*****
 * The <tt>ProjectType</tt> class is used to store min/max information found in the
 * file "Kurse.txt" for a certain project type, e.g. "inf1".
 *****/
class ProjectType {
  public String name;
  public int min, max, slotnumber;
  public List maedchen;

  public TimeInterval incompleteTimeslots;
  
  /** This is just an ordinary constructor. **/
  public ProjectType(String aName, int aMin, int aMax, int aSlotnumber) {
    name = aName;
    min = aMin;
    max = aMax;
    slotnumber = aSlotnumber;
    maedchen = new ArrayList();
    incompleteTimeslots = null;
  } // end of ProjectType(String, int, int, int)
  
} // end of class ProjectType
