/**
 * Implementation of class "GeneticIndividual".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ga;

import helperclasses.*;

import java.io.*;
import java.util.*;

/*****
 * The abstract class <tt>GeneticIndividual</tt> defines the functionality of every
 * individual evolvable by the Genetic Algorithm. It does, for instance, specify
 * mutation, crossover and hill-climbing operators. <br>
 * If one wants to implement a distributed Genetic Algorithm, the methods
 * <tt>getGeneticCode</tt> and <tt>setGeneticCode</tt> must be overridden so that
 * transportation of individuals via an <tt>IslandFerry</tt> object works properly.
 *****/

public abstract class GeneticIndividual implements Comparable {
  /**
   * Objects inheriting the functionality of <tt>GeneticIndividual</tt>
   * <b>must</b> store their current fitness values locally and may only
   * re-compute them if they have changed. This is absolutely necessary
   * for efficient sort operations (remember that in every generation,
   * the entire population is sorted when using RANK_BASED_SELECTION).
   **/
  public final static double NO_FITNESS = Double.NEGATIVE_INFINITY;
  /**
   * Contains the current fitness value of the individual or, in case there
   * is no known fitness value, <tt>NO_FITNESS</tt>.
   **/
  public double fitness;
  
  public GeneticIndividual() {
    fitness = NO_FITNESS;
  }
  
  /**
   * Returns the fitness of the individual. <br>
   * <i>Important:</i> Extensions of <tt>GeneticIndividual</tt> (such as
   * <tt>TTProblem</tt> <b>must</b> store every result of a call in
   * the member variable <tt>fitness</tt>.
   **/
  public abstract double getFitness();

  /** Creates and returns an exact copy of the individual. **/
  public abstract GeneticIndividual duplicate();

  /**
   * Performs a crossover operation on the object itself and the partner individuals
   * specified by <pre>partners</pre>.
   * @param  crossingPoints  determines the number of cp's. zero means uniform cross-over.
   **/
  public abstract GeneticIndividual crossOver(GeneticIndividual[] partners, int crossingPoints);

  /**
   * <tt>mutate(double)</tt> lets the individuals gene sequence mutation with the
   * given mutation rate.
   * @param mutationRate   the probability of a genetic atom to change
   **/
  public abstract void mutate(double mutationRate);

  /**
   * <tt>performDirectedMutation</tt> helps increase the performance of the Genetic Algorithm
   * by directly mutating genes that are likely to be trouble-makers.
   **/
  public abstract void performDirectedMutation(double MutationRate);

  /**
   * The Genetic Algorithm can be forced to perform repair operations
   * on the individuals. This is realized by the <tt>repair</tt>
   * method.
   **/
  public abstract void repair(double repairRate);

  /** Local search, used as an hybrid optimization operator. **/
  public abstract void localSearch();

  /** Forces the individual to do some hill-climbing (e.g. by local search). **/
  public abstract void climb(double climbRate);

  /**
   * This returns a summary of the actual Individual. This could
   * be, for example, a list of all scheduled events.
   **/
  public abstract String getSummary();

  /**
   * The compare method can be used to sort a set of genetic individuals.
   * A call like <tt>Arrays.sort(population);</tt> will result in an array
   * sorted in descending order (i.e. where the individual having the greatest
   * fitness value is on position 0).
   **/
  public final int compareTo(Object o) {
    GeneticIndividual gi = (GeneticIndividual)o;
    if (fitness == NO_FITNESS)
      getFitness();
    if (gi.fitness == NO_FITNESS)
      gi.getFitness();
    if (gi.fitness == fitness)
      return 0;
    if (gi.fitness < fitness)
      return 1;
    return -1;
  } // end of compareTo(Object)

  /** Returns <tt>true</tt> if <tt>compareTo</tt> returns 0, <tt>false</tt> otherwise. **/
  public final boolean equals(GeneticIndividual gi) {
    return (compareTo(gi) == 0);
  }

  /** Produces an XML description of the individual. **/
  public abstract String toXML();

  /** Creates a beutiful HTML view of the individual. **/
  public abstract String toHTML();

  /** Displays some interessant information about the species being evolved. **/
  public abstract String getSpeciesDescription();
  
  /**
   * May be overridden by extending classes in order to provide some protocol
   * information. Timetables (<tt>TTProblem</tt>) can, for example, return statistical
   * information on constraint violation distribution.
   * @return  an <tt>ArrayList</tt> object which contains key-value pairs that can
   * be stored in a table.
   **/
  public List getProtocolInformation() {
    return new ArrayList();
  } // end of getProtocolInformation()

  /**
   * This method returns a bytecode-description of the individual. Classes extending <tt>GeneticIndividual</tt>
   * should override the method in order to be able to migrate between islands in a distributed GA run.
   **/
  public byte[] getGeneticCode() {
    return new byte[0];
  }

  /**
   * This method lets the caller set the bytecode-description of an individual. It is the counterpart to
   * <tt>getGeneticCode()</tt> and should also be overridden.
   **/
  public void setGeneticCode(byte[] data) {
    return;
  }

} // end of class GeneticIndividual



