/**
 * Implementation of class "GlobalMethods".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package helperclasses;

import java.util.*;

/*****
 * This class offers some often-needed methods, such as String->List conversion
 * or backslash-related XML string methods.
 *****/
public final class GlobalMethods {

  /**
   * The method <tt>plainToXML</tt> transform an arbitrary string so that it can
   * be used as an attribute inside an XML file. It does, for example, replace any
   * occurrences of "," by "\," so that the string can be used as part of a comma-separated
   * XML list.
   **/
  public static final String plainToXML(String s) {
    if (s == null) return null;
    s = replaceString(s, "\\", "\\\\");
    s = replaceString(s, ",", "\\,");
    s = replaceString(s, "\"", "\\\"");
    return s;
  } // end of plainToXML(String)

  /** Counterpart to <tt>plainToXML(String)</tt>. **/
  public static final String XMLtoPlain(String s) {
    if (s == null) return null;
    s = replaceString(s, "\\,", ",");
    s = replaceString(s, "\\\"", "\"");
    s = replaceString(s, "\\\\", "\\");
    return s;
  } // end of XMLtoPlain(String)
  
  /** Determines whether a given string is an XML-encoded comma-separated list. **/
  public static final boolean isXMLList(String s) {
    boolean lastWasBackslash = false;
    int length = s.length();
    // a string is an XML list iff it contains a comma withour a preceding backslash
    for (int i = 0; i < length; i++) {
      if (s.charAt(i) == '\\')
        lastWasBackslash = true;
      else {
        lastWasBackslash = false;
        if ((s.charAt(i) == ',') && (!lastWasBackslash))
          return true;
      }
    }
    return false;
  } // end of isXMLList(String)
  
  /** Replaces a substring by another.
   *  @param  s  the string to work on,  old  the old substring,  nu  the new one **/
  public static final String replaceString(String s, String old, String nu) {
    if (s == null) return null;
    String result = "";
    int length = old.length();
    int index = s.indexOf(old);
    while (index >= 0) {
      result += s.substring(0, index) + nu;
      s = s.substring(index + length);
      index = s.indexOf(old);
    }
    return result + s;
  } // end of replaceString(String, String, String)

  /** Transforms a String to a List using the given <tt>separator</tt> char. **/
  public static final ArrayList listFromString(String data, char separator) {
    if (data == null) return null;
    ArrayList result = new ArrayList();
    int index = data.indexOf(separator);
    while (index >= 0) {
      result.add(data.substring(0, index).trim());
      data = data.substring(index + 1);
      index = data.indexOf(separator);
    }
    data = data.trim();
    if (data.length() > 0)
      result.add(data);
    return result;
  } // end of listFromString(String, char)

  /** Transforms a String to a List using the given <tt>separator</tt> string **/
  public static final ArrayList listFromString(String data, String separator) {
    if (data == null) return null;
    ArrayList result = new ArrayList();
    int length = separator.length();
    int index = data.indexOf(separator);
    while (index >= 0) {
      result.add(data.substring(0, index).trim());
      data = data.substring(index + length);
      index = data.indexOf(separator);
    }
    data = data.trim();
    if (data.length() > 0)
      result.add(data);
    return result;
  } // end of listFromString(String, char)

  /** Transforms an XML-encoded comma-separated string list to an ordinary list. **/
  public static final ArrayList listFromXMLString(String data) {
    if (data == null) return null;
    ArrayList result = new ArrayList();
    int length = data.length();
    String element = "";
    for (int i = 0; i < length; i++) {
      char c = data.charAt(i);
      switch (c) {
        case '\\': {
          if (i == length - 1)
            break;
          element += data.charAt(++i);
        }; break;
        case ',': {
          if (element.length() > 0)
            result.add(element);
          element = "";
        }; break;
        default:
          element += c;
      }
    }
    if (element.length() > 0)
      result.add(element);
    return result;
  } // end of listFromXMLString(String)


  /** Used for testing purposes only. **/
  public static void checkTime() {
    System.out.println("Time elapsed: " + (System.currentTimeMillis() - GlobalObjects.lastTimeCheck) + " ms.");
    GlobalObjects.lastTimeCheck = System.currentTimeMillis();
  }

  /** Used for testing purposes only. **/
  public static void checkTime(String stuff) {
    System.out.print(stuff + " --- ");
    checkTime();
  }

  /** Used for testing purposes only. **/
  public static void checkCallCounter() {
    GlobalObjects.callCounter++;
    if ((GlobalObjects.callCounter % 100000) == 0)
      System.out.println("callCounter: " + GlobalObjects.callCounter);
  }

  /** Mixes a list randomly. **/
  public static List mixList(List l) {
    int index;
    ArrayList result = new ArrayList();
    int size = l.size();
    Random random = GlobalObjects.random;
    while (size > 0) {
      index = random.nextInt(size--);
      result.add(l.get(index));
      l.remove(index);
    }
    return result;
  } // end of mixList(List)
  
} // end of class GlobalMethods


