/**
 * Implementation of class "MultiList".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package helperclasses;

import java.util.*;

/*****
 * The <tt>MultiListException</tt> class is used by the <tt>MultiList</tt>
 * class in order to throw some fancy exceptions.
 *****/
class MultiListException extends RuntimeException {
  private String description;

  public MultiListException() {
    description = null;
  }

  public MultiListException(String aDesc) {
    description = aDesc;
  }

  public String toString() {
    if (description == null)
      return "MultiListException";
    else
      return "MultiListException: " + description;
  }

} // end of class MultiListException

/*****
 * Instances of the class <tt>MultiList</tt> are multi-dimensional arrays
 * (1 <= dimension <= 3). The elements inside this array are accesses via
 * the <tt>get</tt> and <tt>set</tt> methods. The size of the array can
 * be modified by calling the <tt>setSize</tt> method.
 *****/
public class MultiList {
  private int dimension;
  private int extension[];
  private List contents;

  /*****
   * Creates new MultiList of the dimension given.
   * @param dim   the number of dimensions of this MultiList
   *****/
  public MultiList(int dim) {
    dimension = dim;
    if ((dimension < 1) || (dimension > 3))
      throw new MultiListException("dimension must be between 1 and 3");
    extension = new int[dimension];
    for (int i = 0; i < dimension; i++)
      extension[i] = 0;
    int cnt = dimension;
    contents = new ArrayList();
  }

  /*****
   * checkDimensionBounds(int, int) is called by all the get(...) methods
   * and throws a MultiListException if <tt>dim</tt> is bigger than the number
   * of dimensions of the MultiList or if <tt>index</tt> is beyond the current
   * bounds.
   *****/
  private void checkDimensionBounds(int index, int dim) {
    if (dimension < dim)
      throw new MultiListException("list has too few dimensions");
  }

  private void setListLen(List l, int len){
    while (l.size() < len)
      l.add(new ArrayList());
    while (l.size() > len)
      l.remove(l.size() - 1);
  }

  /*****
   * The <tt>getDimensionCount()</tt> method returns the number
   * of dimensions this MultiList has provided</tt>.
   *****/
  public int getDimensionCount() {
    return dimension;
  }

  /*****
   * The <tt>get(int)</tt> method returns the element at the specified
   * position in this MultiList in the case <tt>dimension = 1</tt>.
   *****/
  public Object get(int x) {
    checkDimensionBounds(x, 1);
    return contents.get(x);
  }

  /*****
   * The <tt>get(int)</tt> method sets the element at the specified to the
   * parameter <tt>o</tt>.
   * @return the element previously at the specified position
   *****/
  public Object set(int x, Object o) {
    checkDimensionBounds(x, 1);
    return contents.set(x, o);
  }

  /*****
   * Sets the size of the list in the case <tt>dimension = 1</tt>.
   *****/
  public void setSize(int x) {
    if (x < 0)
      throw new MultiListException("negative list size");
    setListLen(contents, x);
    extension[0] = x;
  }

  /*****
   * The method returns the size of dimension 1
   *****/
  public int size() {
      return extension[0];
  }

  /*****
   * The method returns the size of dimension n
   *****/
  public int size(int n) {
    if ((n < 1) || (n > dimension))
      throw new MultiListException("dimension must be between 1 and " + dimension);

    return extension[n-1];
  }

  /*****
   * The <tt>get(int, int)</tt> method returns the element at the specified
   * position in this MultiList in the case <tt>dimension = 2</tt>.
   *****/
  public Object get(int x, int y) {
    checkDimensionBounds(y, 2);
    List list = (List)get(x);
    return list.get(y);
  }

  /*****
   * Pendant of <tt>set(int)</tt> in the case <tt>dimension = 2</tt>.
   *****/
  public Object set(int x, int y, Object o) {
    checkDimensionBounds(y, 2);
    List list = (List)get(x);
    return list.set(y, o);
  }

  /*****
   * Pendant of <tt>setSize(int)</tt> in the case <tt>dimension = 2</tt>.
   *****/
  public void setSize(int x, int y) {
    if (dimension < 2)
      throw new MultiListException("list has too few dimensions");
    if (y < 0)
      throw new MultiListException("negative list size");
    setSize(x);
    for (int cnt = 0; cnt < x; cnt++) {
      List list = (List)get(cnt);
      setListLen(list, y);
    }
    extension[1] = y;
  }

  /*****
   * The <tt>get(int, int, int)</tt> method returns the element at the specified
   * position in this MultiList in the case <tt>dimension = 3</tt>.
   *****/
  public Object get(int x, int y, int z) {
    checkDimensionBounds(z, 3);
    List list = (List)get(x, y);
    return list.get(z);
  }

  /*****
   * Pendant of <tt>set(int)</tt> in the case <tt>dimension = 3</tt>.
   *****/
  public Object set(int x, int y, int z, Object o) {
    checkDimensionBounds(z, 3);
    List list = (List)get(x, y);
    return list.set(z, o);
  }

  /*****
   * Pendant of <tt>setSize(int)</tt> in the case <tt>dimension = 3</tt>.
   *****/
  public void setSize(int x, int y, int z) {
    if (dimension < 3)
      throw new MultiListException("list has too few dimensions");
    if (z < 0)
      throw new MultiListException("negative list size");
    setSize(x, y);
    for (int cnt1 = 0; cnt1 < x; cnt1++)
      for (int cnt2 = 0; cnt2 < y; cnt2++) {
        List list = (List)get(cnt1, cnt2);
        setListLen(list, z);
      }
    extension[2] = z;
  }

  /*****
   * Checks if the specified object is contained in this MultiList.
   * @return true if this MultiList contains the specified element.
   *****/
  public boolean contains(Object o) {
    List xList = contents;
    for (int x = 0; x < extension[0]; x++) {
      if (dimension > 1) {
        List yList = (List)xList.get(x);
        for (int y = 0; y < extension[1]; y++) {
          if (dimension > 2) {
            List zList = (List)yList.get(y);
            for (int z = 0; z < extension[2]; z++)
              if (((Comparable)o).compareTo(zList.get(z)) == 0)
                return true;
          }
          else
            if (((Comparable)o).compareTo(yList.get(y)) == 0)
              return true;
        }
      }
      else
        if (((Comparable)o).compareTo(xList.get(x)) == 0)
          return true;
    }
    return false;
  } // end of MultiList.contains(Object)

} // end of class MultiList

