/**
 * Implementation of class "SortedList".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package helperclasses;

import java.util.*;

/*****
 * Instances of the class <tt>SortedList</tt> implement sorted lists, i.e.
 * containers whose elements can be accessed directly via the <tt>get</tt>
 * and <tt>set</tt> methods, but on which search operations (<tt>contains</tt>
 * and <tt>indexOf</tt>) can be performed with logarithmic complexity.
 *****/
public class SortedList implements List {
  private int size, officialSize;
  private Object elems[];
  private boolean sorted;

  /***
   * An instance of this class will be returned when calling one of the
   * methods <tt>iterator()</tt>, <tt>listIterator</tt>
   **/
  private class SortedListIterator implements ListIterator {

    /** Index of element to be returned by subsequent call to <tt>next</tt>. **/
    int cursor;

    public SortedListIterator() {
      cursor = 0;
    }

    public SortedListIterator(int index) {
      cursor = index;
    }

    /**
     * Index of element returned by most recent call to next or
     * previous.  Reset to -1 if this element is deleted by a call to remove.
     **/
    int lastRet = -1;

    public boolean hasNext() {
      return (cursor < size());
    }

    public boolean hasPrevious() {
      return (cursor > 0);
    }

    public java.lang.Object next() {
      try {
        Object next = get(cursor);
        lastRet = cursor++;
        return next;
      } catch(IndexOutOfBoundsException e) {
        throw new NoSuchElementException();
      }
    } // end of next()

    public java.lang.Object previous() {
      try {
        Object previous = get(--cursor);
        lastRet = cursor;
        return previous;
      } catch(IndexOutOfBoundsException e) {
        throw new NoSuchElementException();
      }
    } // end of previous()

    public void remove() {
      throw new UnsupportedOperationException();
    }

    public int nextIndex() {
      return cursor;
    }

    public int previousIndex() {
      return (cursor - 1);
    }

    public void set(Object o) {
      throw new UnsupportedOperationException();
    }

    public void add(Object o) {
      throw new UnsupportedOperationException();
    }

  } // end of inner class SortedListIterator

  /** This is just an ordinary constructor. **/
  public SortedList() {
    int i;
    size = 8;
    officialSize = 0;
    elems = new Object[size];
    for (i = 0; i < size; i++)
      elems[i] = null;
    sorted = true;
  } // end of SortedList()

  /** Adjusts the size of the internal array. **/
  private void adjustSize(int newSize) {
    int i;
    Object newElems[] = new Object[newSize];
    for (i = 0; i < officialSize; i++)
      newElems[i] = elems[i];
    for (i = officialSize; i < newSize; i++)
      newElems[i] = null;
    elems = newElems;
    size = newSize;
  } // end of adjustSize(int)

  /**
   * Sorts the internal list using the standard mergesort algorithm,
   * which can be found in class <tt>Arrays</tt>.
   **/
  public void sortList() {
    if (!sorted) {
      Arrays.sort(elems, 0, officialSize);
      sorted = true;
    }
  }

  /**
   * Checks if a specific element is a member of the list.
   * @param o   is the element to look for
   * @return true if the element is part of the list, false otherwise
   **/
  public boolean contains(Object o) {
    return (indexOf(o) >= 0);
  }

  /** Returns the size (number of elements) of the list. **/
  public int size() {
    return officialSize;
  }

  /** The <tt>subList(int, int)</tt> method is not implemented. **/
  public List subList(int fromIndex, int toIndex) {
    throw new UnsupportedOperationException();
  }

  /** Returns true if there are no elements in the list, false otherwise. **/
  public boolean isEmpty() {
    return (officialSize == 0);
  }

  /**
   * Returns an <tt>Iterator</tt> that is able to iterate through the list. The object
   * returned is an instance of the private nested class <tt>SortedListIterator</tt>.
   **/
  public Iterator iterator() {
    return new SortedListIterator();
  }

  /**
   * Returns a <tt>ListIterator</tt> that is able to iterate through the list. The object
   * returned is an instance of the private nested class <tt>SortedListIterator</tt>.
   **/
  public ListIterator listIterator() {
    return new SortedListIterator();
  }

  /**
   * The <tt>listIterator(int)</tt> method is not implemented.
   **/
  public ListIterator listIterator(int index) {
    return new SortedListIterator(index);
  }

  /**
   * This method returns the index (position number the the first occurrence) of the object
   * specified by <tt>o</tt> within the list. If <tt>o</tt> is no member of the list,
   * <tt>-1</tt> will be returned.
   **/
  public int indexOf(Object o) {
    if (!sorted) sortList();
    int index, comparison;
    int lowerIndex = 0;
    int upperIndex = officialSize - 1;
    if (officialSize == 0)
      return -1;
    while (true) {
      index = (lowerIndex + upperIndex + 1) / 2;
      comparison = ((Comparable)elems[index]).compareTo(o);
      if (comparison == 0)
        return index;
      if (lowerIndex >= upperIndex)
        return -1;
      if (comparison < 0)
        lowerIndex = index + 1;
      else
        upperIndex = index - 1;
      if (lowerIndex > upperIndex)
        return -1;
    }
  } // end of indexOf(Object)

  /**
   * Returns the last index (maximum position) of the object specified by <tt>o</tt> within
   * the list. If <tt>o</tt> is no member of the list, <tt>-1</tt> will be returned. If there
   * is only one entry in the list pointing to the object specified, <tt>lastIndexOf</tt>
   * will return the same as <tt>indexOf</tt>.
   **/
  public int lastIndexOf(Object o) {
    int index = indexOf(o);
    if (index >= 0) {
      while ((index < officialSize - 1) && (((Comparable)elems[index + 1]).compareTo(o) == 0))
        index++;
      return index;
    }
    return -1;
  }

  /**
   * Removes the object specified by <tt>o</tt> from the list.
   * @param o   the object to be removed
   * @return true, if <tt>o</tt> can be found within the list, false otherwise
   **/
  public boolean remove(Object o) {
    int index = indexOf(o);
    if (index >= 0) {
      remove(index);
      return true;
    }
    return false;
  }

  /**
   * Removes the object that has the position <tt>index</tt>.
   * @return The object that was found at the specified position and then removed from the list.
   **/
  public Object remove(int index) {
    if ((index >= 0) && (index < officialSize)) {
      Object result = elems[index];
      for (int i = index; i < officialSize - 1; i++)
        elems[i] = elems[i + 1];
      officialSize--;
      return result;
    }
    else
      throw new IndexOutOfBoundsException("!(0 <= " + index + " < " + officialSize + ")");
  }

  /**
   * The <tt>add(int, Object)</tt> method is not implemented.
   **/
  public void add(int index, Object element) {
    throw new UnsupportedOperationException();
  }

  /**
   * The <tt>add</tt> method adds a new element to the list.
   * @return This method always returns <tt>true</tt>.
   **/
  public boolean add(Object o) {
    sorted = false;
    if (officialSize >= size)
      adjustSize(size + 8);
    elems[officialSize++] = o;
    return true;
  }

  public Object set(int index, Object element) {
    if ((index >= 0) && (index < officialSize))
      elems[index] = element;
    else
      throw new IndexOutOfBoundsException("!(0 <= " + index + " < " + officialSize + ")");
    sorted = false;
    return null;
  }

  public Object get(String name) {
    int index = indexOf(name);
    if (index >= 0)
      return elems[index];
    else
      return null;
  }

  public Object get(int index) {
    if ((index >= 0) && (index < officialSize))
      return elems[index];
    else
      throw new IndexOutOfBoundsException("!(0 <= " + index + " < " + officialSize + ")");
  }

  public void clear() {
    officialSize = 0;
    adjustSize(8);
    sorted = true;
  }

  /**
   * The <tt>retainAll(Collection)</tt> method is not implemented.
   **/
  public boolean retainAll(Collection c) {
    throw new UnsupportedOperationException();
  }

  public boolean removeAll(Collection c) {
    int oldSize = officialSize;
    Iterator i = c.iterator();
    while (i.hasNext())
      remove(i.next());
    return (oldSize != officialSize);
  }

  public boolean addAll(Collection c) {
    Iterator i = c.iterator();
    while (i.hasNext())
      add(i.next());
    return true;
  }

  /**
   * The <tt>addAll(int, Collection)</tt> method is not implemented.
   **/
  public boolean addAll(int index, Collection c) {
    throw new UnsupportedOperationException();
  }

  public boolean containsAll(Collection c) {
    Iterator i = c.iterator();
    while (i.hasNext())
      if (!contains(i.next()))
        return false;
    return true;
  }

  public Object[] toArray(Object[] arr) {
    int i;
    System.arraycopy(elems, 0, arr, 0, officialSize);
    return arr;
  }

  public Object[] toArray() {
    int i;
    Object arr[] = new Object[officialSize];
    System.arraycopy(elems, 0, arr, 0, officialSize);
    return arr;
  }

  public String toString() {
    String result = "[";
    for (int i = 0; i < officialSize - 1; i++)
      result += elems[i] + ",";
    if (officialSize > 0)
      result += elems[officialSize - 1];
    return result + "]";
  } // end of toString()

} // end of class SortedList

