/**
 * Implementation of class "XMLParser".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package helperclasses;

import java.io.*;
import java.util.*;

/*****
 * The <tt>XMLParser</tt> class enables any TTProblem to initialize its components
 * (events, ressources, constraints) by processing the XML data file belonging to
 * a certain problem. The XML data file is part of the GTL description of a
 * timetabling problem. <br>
 * This class has been quite poorly documented, since it is not part of the
 * actual Timetabling GA. Please take this into consideration when trying to
 * understand the dirty parsing done here.
 *****/
public final class XMLParser {


  // BEGIN: nested classes

  
  /*****
   * The class <tt>XMLData</tt> describes an XML file.
   *****/
  public final class XMLData {
    /** General XML file information. **/
    public String version, encoding;
    /** The XML file's main item. In case of TTProblems, this is "timetable". **/
    public XMLItem item;
  } // end of class XMLData

  /*****
   * XML items have a name, a list of attributes, a list of subitems contained in
   * this item (between the opening and closing tag) and a plain text.
   *****/
  public final class XMLItem {
    public String name;
    public List attributes;
    public List subItems;
    public String text;

    public XMLItem(String name) {
      this.name = name;
      text = "";
      attributes = new ArrayList();
      subItems = new ArrayList();
    }

    /** Returns the value of the attribute specified by <tt>key</tt>. **/
    public String getAttributeValue(String key) {
      for (int i = attributes.size() - 1; i >= 0; i--) {
        XMLAttribute attr = (XMLAttribute)attributes.get(i);
        if (attr.name.equals(key))
          return attr.value;
      }
      return null;
    }

    public String toString() {
      return "<" + name + " " + attributes + ">";
    }
  
  } // end of class XMLItem

  /*****
   * An XML attribute is of the form NAME="VALUE".
   *****/
  public final class XMLAttribute {
    public String name, value;
  
    public XMLAttribute(String name, String value) {
      this.name = name;
      this.value = value;
    }
  
    public String toString() {
      return name + "=\"" + value + "\"";
    }
  
  } // end of class XMLAttribute

   
  // END: nested classes


  private char parseData[];
  private int parseIndex;

  public XMLParser() {
    parseData = new char[1];
    parseData[0] = ' ';
    parseIndex = 0;
  }

  /** Parses the given input file and returns (upon success) a reference to the
    * XML structure created from the input file. **/
  public final XMLData parse(String inputFile) {
    try {
      char data[] = null;
      FileReader reader = new FileReader(inputFile);
      char buffer[] = new char[16384];
      int charsRead, i;
      // first, read the whole file into a char[] object
      while ((charsRead = reader.read(buffer)) > 0) {
        if (data == null) {
          data = new char[charsRead];
          System.arraycopy(buffer, 0, data, 0, charsRead);
        }
        else {
          char newBuffer[] = new char[data.length + charsRead];
          System.arraycopy(data, 0, newBuffer, 0, data.length);
          System.arraycopy(buffer, 0, newBuffer, data.length, charsRead);
          data = newBuffer;
        }
      }
      // then, turn over to parse(char[])
      return parse(data);
    }
    catch (IOException io) {
      System.err.println("Exception in XMLParser.parse(String):");
      System.err.println("  " + io.toString());
      return null;
    }
  } // end of parse(String)
  
  /** Counterpart to <tt>parse(String)</tt> that takes directly the string
    * data to be parsed. **/
  public final XMLData parse(char data[]) {
    XMLData result = new XMLData();
    String element;
    parseData = data;
    parseIndex = 0;
    element = getNextElement();
    if ((!element.startsWith("<?xml ")) || (!element.endsWith("?>")))
      return null;
    element = element.substring(5, element.length() - 2);
    List keyValues = getKeyValuePairs(element);
    for (int i = 0; i < keyValues.size(); i++) {
      XMLAttribute attr = (XMLAttribute)keyValues.get(i);
      if (attr.name.equals("version"))
        result.version = (String)attr.value;
      if (attr.name.equals("encoding"))
        result.encoding = (String)attr.value;
    }
    result.item = parse();
    return result;
  } // end of parse(char[])

  /** Returns the list of attributes of an element as key-value pairs. **/
  private final List getKeyValuePairs(String attributes) {
    List result = new ArrayList();
    attributes = attributes.trim();
    String name, value;
    int index, index2;
    attributes = GlobalMethods.replaceString(attributes, "= \"", "=\"");
    index = attributes.indexOf("=\"");
    while (index > 0) {
      name = attributes.substring(0, index).trim();
      attributes = attributes.substring(index + 2);
      index2 = attributes.indexOf('"');
      if (index2 < 0) return result;
      value = attributes.substring(0, index2);
      attributes = attributes.substring(index2 + 1).trim();
      result.add(new XMLAttribute(name, value));
      index = attributes.indexOf("=\"");
    }
    return result;
  } // end of getKeyValuePairs(String)
  
  /** Parses an XML item, e.g. &lt;Event name="hallo"&gt;Blablabla &lt;Zeug/&gt; &lt;/Event&gt;. **/
  private final XMLItem parse() {
    String element;
    int index;
    boolean finished;
    element = getNextElement();
    if (element.charAt(0) == '<') {
      XMLItem item = null;
      if (element.endsWith("/>"))
        finished = true;
      else
        finished = false;
      index = element.indexOf(' ');
      if (index < 0)
        item = new XMLItem(element.substring(1, element.length() - 1));
      else {
        item = new XMLItem(element.substring(1, index));
        element = element.substring(index, element.length() - 1).trim();
        item.attributes = getKeyValuePairs(element);
      }
      while (!finished) {
        if (nextChar() == '<') {  // sub-item
          int saveIndex = parseIndex;
          element = getNextElement();
          if (element.equals("</" + item.name + ">")) {
            finished = true;
          }
          else {
            parseIndex = saveIndex;
            item.subItems.add(parse());
          }
        }
        else // ordinary text
          item.text += getNextElement() + "\n";
      };
      return item;
    }
    return null;
  } // end of parse()
  
  private final void skipSpaces() {
    while ((parseIndex < parseData.length) && (parseData[parseIndex] <= ' '))
      parseIndex++;
  }
  
  private char nextChar() {
    int myIndex = parseIndex;
    while ((myIndex < parseData.length) && (parseData[myIndex] <= ' '))
      myIndex++;
    if (myIndex >= parseData.length)
      return 0;
    else
      return parseData[myIndex];
  } // end of nextChar()

  /** Returns the next tag or text area. **/
  private final String getNextElement() {
    String result;
    int myIndex = parseIndex;
    if (nextChar() == '<') {  // next element is an XML tag :-)
      boolean inQuotation = false;
      boolean lastWasBackslash = false;
      boolean finished = false;
      skipSpaces();
      do {
        if (parseData[myIndex] == '"') {
          if (!inQuotation)
            inQuotation = true;
          else
            if (!lastWasBackslash)
              inQuotation = false;
        }
        if ((parseData[myIndex] == '>') && (!inQuotation))
          finished = true;
        if (parseData[myIndex] == '\\')
          lastWasBackslash = true;
        else
          lastWasBackslash = false;
        myIndex++;
      } while (!finished);
      result = new String(parseData, parseIndex, myIndex - parseIndex);
      parseIndex = myIndex;
    }
    else {  // next element is some ordinary text :-(
      while ((myIndex < parseData.length) && (parseData[myIndex] != '<'))
        myIndex++;
      result = new String(parseData, parseIndex, myIndex - parseIndex);
      parseIndex = myIndex;
    }
    return result;
  } // end of getNextElement()

} // end of class XMLParser

