/**
 * Implementation of class "PlanningClass".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;
import java.lang.reflect.*;

/*****
 * <tt>PlanningClass</tt> is the super-class of all relevant objects
 * related to a scheduling/time-tabling problem: events and resources.
 *****/
public abstract class PlanningClass implements Comparable {
  /** Internal value for updating self-managed lists. **/
  protected static final int LIST_GRANULARITY = 16;

  /** Every object (resource, event) needs a unique name. **/
  public String name;
  /** Every object can hold a list of (hard and soft) constraints that shall not be violated. **/
  private Constraint constraints[];
  private int constraintCount;
  /** <tt>owner</tt> is the <tt>TTProblem</tt> instance that the Planning Class belongs to. **/
  protected TTProblem owner;
  /** An additional description that is supposed to help identify the object. **/
  public String description;
  private String toStringCache;

  /** This is an ordinary constructor. **/
  public PlanningClass() {
    name = "(no name)";
    owner = null;
    constraintCount = 0;
    constraints = new Constraint[LIST_GRANULARITY];
    description = null;
    toStringCache = null;
  } // end of PlanningClass()

  /** Sets the owning <tt>TTProblem</tt> instance. **/
  public void setOwner(TTProblem owner) {
    this.owner = owner;
  }
  
  /** Returns the owning <tt>TTProblem</tt> instance. **/
  public TTProblem getOwner() {
    return owner;
  }

  /** Sets the description text. **/
  public void setDescription(String aDesc) {
    description = aDesc;
  }

  /** This method lets the caller add a new constraint to the planning class' constraint list. **/
  public final void addConstraint(Constraint c) {
    if (c == null) return;
    if (constraintCount >= constraints.length) {
      Constraint newConstraints[] = new Constraint[constraints.length + LIST_GRANULARITY];
      for (int i = 0; i < constraints.length; i++)
        newConstraints[i] = constraints[i];
      constraints = newConstraints;
    }
    constraints[constraintCount++] = c;
  } // end of addConstraint(Constraint)

  /** Returns a list of all constraints belonging to this planning class. **/
  public final ArrayList getConstraints() {
    ArrayList result = new ArrayList(constraintCount);
    for (int i = 0; i < constraintCount; i++)
      result.add(constraints[i]);
    return result;
  } // end of getConstraints()

  /** Returns a list of <tt>ConstraintViolation</tt> instances describing the violations caused by <tt>this</tt>. **/
  public final ArrayList computeConstraintViolations() {
    ConstraintViolation cv;
    ArrayList result = new ArrayList();

    for (int i = constraintCount - 1; i >= 0; i--) {
      try {
        cv = constraints[i].compute(this);
        if (cv != null)
          result.add(cv);
      } catch (Exception e) {
        System.err.println("Error computing Constraint " + constraints[i].name);
        System.err.println("Exception " + e);
      }
    }
    return result;
  } // end of computeConstraintViolations()

  /** Returns the number of constraint violations caused by <tt>this</tt>. **/
  protected final int getViolationCount() {
    ArrayList l = computeConstraintViolations();
    return l.size();
  }

  /** Returns the penalty points generated by <tt>this</tt>. **/
  protected double getPenaltyPoints() {
    ConstraintViolation cv;
    double result = 0;

    for (int i = constraintCount - 1; i >= 0; i--) {
      try {
        cv = constraints[i].compute(this);
        if (cv != null)
          result += cv.penalty;
      } catch (Exception e) {
        System.err.println("Error computing Constraint " + constraints[i].name);        System.err.println("Exception " + e);
      }
    }
    return result;
  } // end pf getPenaltyPoints()

  /**
   * <tt>setVariable</tt> is used by the initialization routine of the TTProblem instance.
   * Since instantiation data is defined in an XML file in the form features="servus,ade",
   * all planning classes must offer a possibility to dynamically select variables
   * by their names. If XML instantiation is wished, this methodhas to be overridden.
   **/
  public void setVariable(String key, String value) {
    if (key.equals("name"))
      name = GlobalMethods.XMLtoPlain(value);
    else if (key.equals("description"))
      description = GlobalMethods.XMLtoPlain(value);
    else if (key.equals("constraints")) {
      constraintCount = 0;
      constraints = new Constraint[LIST_GRANULARITY];
      List list = GlobalMethods.listFromXMLString(value);
      for (int i = 0; i < list.size(); i++)
        addConstraint(owner.getConstraint((String)list.get(i)));
    } else {
      // user defined member -> use Java reflection mechanisms
      Class c = this.getClass();
      Field[] f = c.getDeclaredFields(); // get all members declared
      int i = 0;

      while(i < f.length) {
        if(!key.equals(f[i].getName())) {
           i++;
           continue;
        }

        // correct member found
        // get class description of this member
        Class memberClass = f[i].getType();

        try {
          if (memberClass.isPrimitive()) {
            if (memberClass.equals(java.lang.Boolean.TYPE)) {
              f[i].setBoolean(this, Boolean.getBoolean(value));
              return;
            }
            else if (memberClass.equals(java.lang.Short.TYPE)) {
              f[i].setShort(this, Short.parseShort(value));
              return;
            }
            else if (memberClass.equals(java.lang.Integer.TYPE)) {
              f[i].setInt(this, Integer.parseInt(value));
              return;
            }
            else if (memberClass.equals(java.lang.Long.TYPE)) {
              f[i].setLong(this, Long.parseLong(value));
              return;
            }
            else if (memberClass.equals(java.lang.Float.TYPE)) {
              f[i].setFloat(this, Float.parseFloat(value));
              return;
            }
            else if (memberClass.equals(java.lang.Double.TYPE)) {
              f[i].setDouble(this, Double.parseDouble(value));
              return;
            }
            else {
              System.err.println("Error in definition:\nPrimitive type "+ key + " found in " + c.getName() + " is not supported!");
              return;
            }
          } else {
            // not primitive type

            // single String?
            if (memberClass.getName().equals("java.lang.String")) {
              f[i].set(this, value);
              return;
            }
            else if (memberClass.getName().equals("helperclasses.MultiList")) {
              // get concrete MultList object
              MultiList m = (MultiList)f[i].get(this);
              if (m.getDimensionCount() != 1) {
                System.err.println("Dimensions greater than one are currently not supported: " + memberClass.getName() + " in declaration of " + c.getName());
                return;
              }
              List list = GlobalMethods.listFromXMLString(value);
              m.setSize(list.size());
              for (int listCount = 0; listCount < list.size(); listCount++)
                m.set(listCount, (String) list.get(listCount));
              return;
            }
            else if ((memberClass.getName().equals("java.util.List")) ||
                     (memberClass.getName().equals("java.util.ArrayList"))) {
              List list = (List)f[i].get(this);
              list.clear();
              List newList = GlobalMethods.listFromXMLString(value);
              list.addAll(newList);
              return;
            }
            else {
              System.err.println("Unkown type " + memberClass.getName() + " in declaration of " + c.getName());
              return;
            }
          }
        } catch (IllegalAccessException e) { System.out.println(e);}

        i++;
      }

      // key that has not been declared as a member
      System.err.println("Error in XML file:\nKey "+ key + " has not been declared as member of " + c.getName());
    }
  } // end of setVariable(String, String)

  /** Returns the name of the object. **/
  public String toString() {
    if (toStringCache == null) {
      if (description == null)
        toStringCache = name;
      else
        toStringCache = name + " (" + description + ")";
    }
    return toStringCache;
  } // end of toString()

  /** Compares to objects by their name. Hence, a call to <tt>compareTo(o)</tt> is equivalent to
      <tt>name.compareTo(o)</tt>. **/
  public int compareTo(Object o) {
    if (o == this) return 0;
    if (o instanceof String)
      return name.compareTo(o);
    return name.compareTo(((PlanningClass)o).name);
  } // end of compareTo(Object)

  /** Returns true if <tt>compareTo</tt> states equality (returns zero). **/
  public boolean equals(Object o) {
    return (compareTo(o) == 0);
  }

} // end of class PlanningClass

