/**
 * Implementation of class "Resource".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;

/*****
 * All resources (teachers, rooms, ...) that appear in any timetabling problem (such as school timetabling,
 * employee scheduling, ...) are extensions of this class.
 *****/
public abstract class Resource extends PlanningClass {
  /**
   * Resources are <i>in use</i> by an event. In order to keep the
   * different views consistent, every resource has got this list of
   * all events that it is involved in.
   **/
  protected Assignment assignments[];
  protected int assignmentCount;
  /**
   * Resources can have certain features, e.g. teachers can have the
   * ability to teach certain subjects. Since this happens quite often,
   * we decided to let <tt>features</tt> be an attribute of all <tt>Resource</tt>
   * (and derived classes') instances.
   **/
  protected String features[];
  protected int featureCount;

  /** Just an ordinary constructor. **/
  public Resource() {
    assignments = new Assignment[LIST_GRANULARITY]; assignmentCount = 0;
    features = new String[LIST_GRANULARITY]; featureCount = 0;
  }

  /** See class <tt>PlanningClass</tt> for details. **/
  public void setVariable(String key, String value) {
    if (key.equals("features")) {
      featureCount = 0;
      features = new String[LIST_GRANULARITY];
      List list = GlobalMethods.listFromXMLString(value);
      for (int i = 0; i < list.size(); i++)
        addFeature((String)list.get(i));
    }
    else
      super.setVariable(key, value);
  } // end of setVariable(String, String)

  private final int getIndexOfEvent(Event e) {
    for (int i = 0; i < assignmentCount; i++)
      if (assignments[i].event == e)
        return i;
    return -1;
  }

  private final int getIndexOfAssignment(Assignment ass) {
    for (int i = 0; i < assignmentCount; i++)
      if (assignments[i] == ass)
        return i;
    return -1;
  }

  protected void clearAssignments() {
    assignmentCount = 0;
  }

  /** Adds <tt>ass</tt> to the resource's assignment list. **/
  protected void addAssignment(Assignment ass) {
    if (assignmentCount >= assignments.length) {
      Assignment newAssignments[] = new Assignment[assignmentCount + LIST_GRANULARITY];
      for (int i = 0; i < assignmentCount; i++)
        newAssignments[i] = assignments[i];
      assignments = newAssignments;
    }
    assignments[assignmentCount++] = ass;
  } // end of addAssignment(Assignment)

  /** Removes <tt>ass</tt> from the resource's assignment list. **/
  protected void removeAssignment(Assignment ass) {
    int i, index;
    for (index = 0; index < assignmentCount; index++) {
      if (assignments[index] == ass) {
        for (i = index; i < assignmentCount - 1; i++)
          assignments[i] = assignments[i + 1];
        assignmentCount--;
        return;
      }
    }
  } // end of removeAssignment(Assignment)

  /** Returns a list of all events that this resource has been assigned to. **/
  public final List getEvents() {
    ArrayList result = new ArrayList();
    for (int i = 0; i < assignmentCount; i++)
      result.add(assignments[i].event);
    return result;
  }

  /** Returns a list of all events that this resource has been assigned to. **/
  public final ArrayList getAssignments() {
    ArrayList result = new ArrayList();
    for (int i = 0; i < assignmentCount; i++)
      result.add(assignments[i]);
    return result;
  }

  public final int getAssignmentCount() {
    return assignmentCount;
  }

  /** Adds the feature to this resource. **/
  public final void addFeature(String feature) {
    if (featureCount >= features.length) {
      String newFeatures[] = new String[featureCount + LIST_GRANULARITY];
      for (int i = 0; i < featureCount; i++)
        newFeatures[i] = features[i];
      features = newFeatures;
    }
    features[featureCount++] = feature;
  }

  /** Returns a list of all features of this resource. **/
  public final List getFeatures() {
    ArrayList result = new ArrayList();
    for (int i = 0; i < featureCount; i++)
      result.add(features[i]);
    return result;
  }

  /** Returns true if <tt>feature</tt> is a member of the <tt>features</tt> array, false otherwise. **/
  public final boolean hasFeature(String feature) {
    for (int i = 0; i < featureCount; i++)
      if (feature.equals(features[i])) return true;
    return false;
  }

  /** Returns a list of all TimeSlots during which this resource is occupied. **/
  public final List getTimeSlots() {
    ArrayList result = new ArrayList();
    int cnt, size, j;
    for (j = 0; j < assignmentCount; j++) {
      // loop through all events and add all timeslots to the list
      Event e = assignments[j].event;
      ResourceList resList = e.getResourceList("TimeSlots");
      if (resList != null) {
        Resource[] list = resList.getAssignmentArray();
        size = list.length;
        for (cnt = 0; cnt < size; cnt++)
          result.add(list[cnt]);
      }
    }
    return result;
  } // end of getTimeSlots()

  /** Returns the XML description of the Resource element. **/
  public String toXML(String type) {
    int j, k;
    String result = "<Resource name=\"" + GlobalMethods.plainToXML(name) +
                    "\" type=\"" + GlobalMethods.plainToXML(type) + "\"";
    result += " features=\"";
    for (j = 0; j < featureCount; j++) {
      result += GlobalMethods.plainToXML(features[j]);
      if (j < featureCount - 1)
        result += ",";
    }
    result += "\" ";
    result += "events=\"";
    for (j = 0; j < assignmentCount; j++) {
      result += GlobalMethods.plainToXML(assignments[j].event.name);
      if (j < assignmentCount - 1)
        result += ",";
    }
    result += "\"";
    if (description == null)
      result += "/>\n";
    else
      result += " description=\"" + GlobalMethods.plainToXML(description) + "\"/>\n";
    return result;
  }

} // end of Resource

