/**
 * Implementation of class "ResourceList".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;

/*****
 * The <tt>ResourceList</tt> class manages the resources used within an
 * <tt>Event</tt> instance. For example, in a highschool timetabling situation,
 * the <tt>ResourceList</tt> belonging to a lesson's teacher would be like this:
 * <tt>new ResourceList("Teacher", 1, 1, teacherList)</tt>.
 *****/
public final class ResourceList implements Comparable {
  private Event owner;  // every ResourceList instance is owned by an event
  private Random random;
  private String name;  // and has got a unique name
  private int min, max; // minimum and maximum assignment count (constraint)
  private int assigned, size;   // number of resources currently assigned and
                               // number of managed resources at all
  private boolean assignment[];  // assignment vector
                        // true means: this resource is currently assigned to owner
                        // false means: it is currently not assigned to owner
  private SortedList takeFrom;  // list of resource components to choose from for assignment
  private Resource takeFromCopy[];  // for faster access
  private Resource assignmentArray[];  // list of all currently assigned resources (for speedup) or NULL
  private boolean fixed; // this flag is set if the assignment is predefined
      // there are two types of fixed resource lists:
      // (a) the list is completely fixed, i.e. there will never be any mutation (this is true if size=min=max)
      // (b) only the value of assigned is fixed (this is true if size!=min=max)
      // if min!=max, then the resource list can be defined fixed

  private Assignment assignmentTuple; // Will be used to keep resource assignment views consistents.
            // As an assignment consists of an event and an associated resource list,
            // <tt>assignmentTuple</tt> contains <tt>owner</tt> and <tt>this</tt>.

  private static final int ONLY_ONE_OPERATION = 0;   // shall the mutation operator consist of several
  private static final int MULTIPLE_OPERATIONS = 1;  // basic operations or just one atomic one?
  private static final int MUTATION_MODE = ONLY_ONE_OPERATION;

  private String extra;  // This variable can be used to store additional information
                         // e.g. something that extends the pure min/max information by defining a preferred assignment count

  /**
   * Creates a new <tt>ResourceList</tt> instance.
   * @param name   The name of this resource ("TimeSlots" for time resource).
   * <br> min   The minimum number of actual resource components assigned
   *            within the resource list.
   * <br> max   The maximum number.
   * <br> takeFrom   The list to choose the components from.
   **/
  public ResourceList(Event owner, String name, int min, int max, List takeFrom) {
    int i;
    assignmentTuple = new Assignment(owner, this);
    this.owner = owner;
    this.name = name;
    this.min = min;
    this.max = max;
    assigned = 0;
    size = takeFrom.size();
    assignment = new boolean[size];
    for (i = 0; i < size; i++)
      assignment[i] = false;
    this.takeFrom = new SortedList();
    takeFromCopy = new Resource[size];
    for (i = 0; i < size; i++)
      this.takeFrom.add(takeFrom.get(i));
    this.takeFrom.sortList();
    for (i = 0; i < size; i++)
      takeFromCopy[i] = (Resource)this.takeFrom.get(i);
    this.random = GlobalObjects.random;
    this.fixed = false;

    if ((min < 0) || (max > size)) {
      // if these natural constraints are violated, correct that
      if (min < 0) min = 0;
      if (max > size) max = size;
    }

    extra = null;
    assignmentArray = null;
  } // end of ResourceList(Event, String, int, int, List)

  /** Creates a new <tt>ResourceList</tt> with fixed assignment. **/
  public ResourceList(Event owner, String name, List takeFrom) {
    this(owner, name, takeFrom.size(), takeFrom.size(), takeFrom);
    fixed = true;
    for (int i = 0; i < size; i++)
      assignResource(i);
  } // end of ResourceList(Event, String, List)

  /**
   * Alternative constructor, to be used when there is only one possible Resource.
   * @param name   The name of this resource ("TimeSlots" for time resource).
   * <br> takeFrom   The resource.
   **/
  public ResourceList(Event owner, String name, Resource takeFrom) {
    assignmentTuple = new Assignment(owner, this);
    this.owner = owner;
    this.name = name;
    this.min = this.max = 1;
    assigned = 0;
    size = 1;
    assignment = new boolean[size];
    this.takeFrom = new SortedList();
    takeFromCopy = new Resource[1];
    this.takeFrom.add(takeFrom);
    this.takeFrom.sortList();
    takeFromCopy[0] = takeFrom;
    this.random = GlobalObjects.random;
    this.fixed = true;
    assignResource(0);
    assignmentArray = null;
  } // end of ResourceList(Event, String, Resource)

  /** Lets the caller fix the resource list. This is only possible if <tt>min == max</tt>. Fixing a resource list
    * implies that the value of the private member variable <tt>assigned</tt> will be constant. **/
  public final void fix() {
    if ((min == max) && (min <= size)) {
      // we can only be fixed if min = max (because otherwise fixing does not make any sense at all)
      fixed = true;
      if (min == size) // everything will be fixed now
        for (int i = 0; i < size; i++)
          assignResource(i);
      else {  // only fix the assignment count
        int index, cnt;
        while (assigned < min) {
          index = random.nextInt(size - assigned);
          cnt = 0;
          while ((assignment[cnt]) || (index > 0)) {
            if (!assignment[cnt]) index--;
            cnt++;
          }
          assignResource(cnt);
        }
        while (assigned > min) {
          index = random.nextInt(assigned);
          cnt = 0;
          while ((!assignment[cnt]) || (index > 0)) {
            if (assignment[cnt]) index--;
            cnt++;
          }
          freeResource(cnt);
        }
      }
    }
  } // end of fix()

  /**
   * Adds the member of <tt>takeFrom</tt> with index number <tt>index</tt> to
   * the list of assigned resource components and updates its view.
   **/
  public final void assignResource(int index) {
    if (assignment[index]) return;
    assignmentArray = null;
    assigned++;
    assignment[index] = true;
    takeFromCopy[index].addAssignment(assignmentTuple);
  }

  /** Similar to <tt>assignResource(int)</tt>. **/
  public final void assignResource(Resource res) {
    int index = takeFrom.indexOf(res);
    if ((index < 0) || (assignment[index])) return;
    assignmentArray = null;
    assigned++;
    assignment[index] = true;
    res.addAssignment(assignmentTuple);
  }
  
  /** Assigns a list of resources to the resource list event. **/
  public final void assignList(List resources) {
    for (int i = resources.size() - 1; i >= 0; i--)
      assignResource((Resource)resources.get(i));
  }

  /** Randomly assigns a previously not assigned resource. **/
  public final void assignRandom() {
    if (assigned < size) {
      int index = random.nextInt(size - assigned);
      int cnt = 0;
      while ((assignment[cnt]) || (index > 0)) {
        if (!assignment[cnt]) index--;
        cnt++;
      }
      assignResource(cnt);
    }
  } // end of assignRandom()

  /** Counterpart to <tt>assignResource(int)</tt>. **/
  public final void freeResource(int index) {
    if (!assignment[index]) return;
    assignmentArray = null;
    assigned--;
    assignment[index] = false;
    takeFromCopy[index].removeAssignment(assignmentTuple);
  }

  /** Counterpart to <tt>assignResource(Resource)</tt>. **/
  public final void freeResource(Resource res) {
    int index = takeFrom.indexOf(res);
    if ((index < 0) || (!assignment[index])) return;
    assignmentArray = null;
    assigned--;
    assignment[index] = false;
    res.removeAssignment(assignmentTuple);
  }

  /** Frees all currently assigned resources. **/
  public final void freeAllResources() {
    for (int i = 0; i < size; i++)
      if (assignment[i])
        freeResource(i);
  }
  
  /** Randomly releases a previously assigned resource. **/
  public final void freeRandom() {
    if (assigned > 0) {
      int index = random.nextInt(assigned);
      int cnt = 0;
      while ((!assignment[cnt]) || (index > 0)) {
        if (assignment[cnt]) index--;
        cnt++;
      }
      freeResource(cnt);
    }
  } // end of freeRandom()
  
  /**
   * This can be used to check if a certain instance is currently assigned.
   * @return true if the resource at position <tt>index</tt> has been assigned,
   * false otherwise.
   **/
   public final boolean isAssigned(int index) {
     return assignment[index];
   }
   
   /** Addition to <tt>isAssigned(int)</tt>. Works on an object directly. **/
   public final boolean isAssigned(Resource res) {
     int index = takeFrom.indexOf(res);
     if (index < 0)
       return false;
     return assignment[index];
   }

  /**
   * Returns the boolean assignment vector. <br>
   * <em>The bit vector returned is the very one the resource list itself works
   * on. So, please be cautious with what you do with it.</em>
   **/
  public final boolean[] getAssignment() {
    return assignment;
  }

  /** Sets the assignment vector to the parameter's value. **/
  public final void setAssignment(boolean source[]) {
    for (int i = 0; i < size; i++) {
      if ((source[i]) && (!(assignment[i])))
        assignResource(i);
      if ((!(source[i])) && (assignment[i]))
        freeResource(i);
    }
  } // end of setAssignment(boolean[])

  /**
   * A more general version of <tt>setAssignment(boolean[])</tt> which
   * is intended to speed up data transfer between <tt>TTChromosome</tt>
   * and <tt>ResourceList</tt>.
   **/
  protected final void setAssignment(boolean source[], int offset) {
    for (int i = 0; i < size; i++) {
      if ((source[offset]) && (!(assignment[i])))
        assignResource(i);
      if ((!(source[offset])) && (assignment[i]))
        freeResource(i);
      offset++;
    }
  } // end of setAssignment(boolean[], int)

  /** Returns a list of all resource components currently assigned. **/
  public final ArrayList getAssignmentList() {
    ArrayList result = new ArrayList(assigned);
    for (int i = 0; i < size; i++)
      if (assignment[i])
        result.add(takeFromCopy[i]);
    return result;
  }

  /**
   * Does the same as <tt>getAssignmentList</tt> except that it returns
   * an array instead of a <tt>List</tt> object.
   **/
  protected final Resource[] getAssignmentArray() {
    if (assignmentArray == null) {
      assignmentArray = new Resource[assigned];
      int index = 0;
      for (int i = 0; i < size; i++)
        if (assignment[i])
          assignmentArray[index++] = takeFromCopy[i];
    }
    return assignmentArray;
  } // end of getAssignmentArray()
    
  
  /** Returns a list of all resource components that are currently free. **/
  public final List getFreeResources() {
    ArrayList result = new ArrayList(size - assigned + 2);
    for (int i = 0; i < size; i++)
      if (!assignment[i])
        result.add(takeFromCopy[i]);
    return result;
  }

  /** Returns the list of all plannable resource components (free and assigned). **/
  public final List getAllResources() {
    ArrayList result = new ArrayList();
    for (int i = 0; i < size; i++)
      result.add(takeFromCopy[i]);
    return result;
  }

  /** Returns the number of resource components currently assigned. **/
  public int assignedCount() {
    return assigned;
  }

  /** Number of resource components managed by this list at all. **/
  public int size() {
    return size;
  }

  /** Returns the Resource instance at position <tt>index</tt>. **/
  public Resource getResource(int index) {
    return takeFromCopy[index];
  }

  /**
   * Returns the <tt>index</tt>'th assigned resource component. <br>
   * <i>Please do not call this method subsequently, since it is <b>really</b>
   * slow (quadratic complexity). Instead, call <tt>getAssignmentList</tt> once
   * and then iterate through the list returned.</i>
   **/
  public Resource getAssignedResource(int index) {
    int count = -1;
    for (int i = 0; i < size; i++) {
      if (assignment[i])
        count++;
      if (count == index)
        return takeFromCopy[i];
    }
    return null;
  } // end of getAssignedResource(int)

  /**
   * Returns the <tt>index</tt>'th free resource component. <br>
   * <i>Please do not call this method subsequently, since it is <b>really</b>
   * slow (quadratic complexity). Instead, call <tt>getFreeResources</tt> once
   * and then iterate through the list returned.</i>
   **/
  public final Resource getFreeResource(int index) {
    int count = -1;
    for (int i = 0; i < size; i++) {
      if (!assignment[i])
        count++;
      if (count == index)
        return takeFromCopy[i];
    }
    return null;
  } // end of getFreeResource(int)


  /**
   * This method is called by <tt>Event.updateViews</tt> and manages
   * the update process for all assigned resource components within the list.
   **/
  public final void updateViews() {
    assigned = 0; assignmentArray = null;
    for (int i = 0; i < size; i++)
      if (assignment[i]) {
        assigned++;
        takeFromCopy[i].addAssignment(assignmentTuple);
      }

    if ((fixed) && (assigned != min))
      // maybe, something went wrong during uniform crossover (can destroy min-max relations when fixed)
      // so, we have to fix this resource list again
      fix();
  } // end of updateViews()

  /** <tt>mixList(List)</tt> runs a simple permutation on a given list. **/
  private final List mixList(List l) {
    ArrayList result = new ArrayList();
    int size = l.size();
    while (size > 0) {
      Object element = l.get(random.nextInt(size--));
      l.remove(element);
      result.add(element);
    }
    return result;
  } // end of mixList(List)

  /** <tt>localSearch()</tt> performs local search in order to minimize penalty points. **/
  public final void localSearch() {
    double penalty;
    double bestPenalty = owner.getPenaltyPoints();

    if ((fixed) && (min == size))  // nothing to do in this case
      return;
    if ((bestPenalty <= 0) || (size == 0))  // only optimize if there is something to optimize
      return;

    if (fixed) {
      if (min == 1) { // this is the easiest case: just try every resource subsequently
        int i, bestChoice;
        bestChoice = -1;
        for (i = 0; i < size; i++)
          if (assignment[i]) {
            freeResource(i);
            bestChoice = i;
            break;
          }
        assignResource(0);
        penalty = owner.getPenaltyPoints();
        if ((bestChoice < 0) || (penalty < bestPenalty)) {
          bestPenalty = penalty;
          bestChoice = 0;
        }
        for (i = 1; i < size; i++) {
          // cycle through takeFrom list and try every resource in that list
          freeResource(i - 1);
          assignResource(i);
          penalty = owner.getPenaltyPoints();
          if (penalty < bestPenalty) {
            bestPenalty = penalty;
            bestChoice = i;
          }
        } // end of loop
        freeResource(size - 1);
        assignResource(bestChoice);
      }  // end of easy case
      else {  // try to exchange resources for the general case with min > 1
        boolean old[] = new boolean[size];
        System.arraycopy(assignment, 0, old, 0, size);
        for (int i = 0; i < 16; i++) {
          assignRandom(); freeRandom();
          penalty = owner.getPenaltyPoints();
          if (penalty < bestPenalty) {
            bestPenalty = penalty;
            System.arraycopy(assignment, 0, old, 0, size);
          }
          else
            setAssignment(old);
        }
      }
    }
    else {  // the resource list has not been fixed
      if (size <= 4) {
        // if the list is not too long, we can try exhaustive search
        boolean bestAssignment[] = new boolean[size];
        System.arraycopy(assignment, 0, bestAssignment, 0, size);
        freeAllResources();
        // only perform full optimization if the assignment vector is not too large
        while (nextAssignment()) {
          penalty = owner.getPenaltyPoints();
          if (penalty < bestPenalty) {
            // if this combination is better than the previous best one, save it
            bestPenalty = penalty;
            System.arraycopy(assignment, 0, bestAssignment, 0, size);
          }
        }
        setAssignment(bestAssignment);
      }
      else {  // only try bit-flipping and hope that helps
        double probability = (double)32 / (double)size;
        int start, end, increment;
        // because it is boring to always walk in the same direction:
        if (random.nextBoolean()) {
          start = 0; end = size; increment = +1;
        }
        else {
          start = size - 1; end = -1; increment = -1;
        }
        for (int i = start; i != end; i += increment) if (random.nextDouble() < probability) {
          if (assignment[i]) freeResource(i);
            else assignResource(i);
          penalty = owner.getPenaltyPoints();
          if (penalty <= bestPenalty)  // if this combination is better, save it
            bestPenalty = penalty;
          else {  // if it is worse, undo changes
            if (assignment[i]) freeResource(i);
              else assignResource(i);
          }
        }
      }  // end of bit-flipping
    }
  } // end of localSearch()

  /**
   * One can loop though all possible combinations of resource assignment vectors
   * by subsequently calling this method.
   * @return  false, if there is no more possible assignment combination <br>
   *          true otherwise
   **/
  private boolean nextAssignment() {
    int index = 0;
    if (assignment[index]) freeResource(index);
      else assignResource(index);
    while (!assignment[index]) {  // bit-flip 1=>0 -> forward carry
      index++;
      if (index >= size)
        return false;
      if (assignment[index]) freeResource(index);
        else assignResource(index);
    }
    return true;
  } // end of nextAssignment(boolean [])

  /**
   * Since resource lists are part of a chromosome, they must offer the
   * ability to mutate. Mutation is done by randomly adding/removing
   * resource components to/from the assignment list.
   **/
  public void mutate() {
    List free, nonfree;

    // in case of an empty takeFrom list, do nothing
    if (size == 0) return;
    
    if (fixed) {  // this resource list has in some way been fixed
      if (min != size) {  // this is a resource list with fixed assignment count but no fixed assignments
        int index, cnt;
        if (MUTATION_MODE == ONLY_ONE_OPERATION) {  // exchange just one resource
          // exchange one resource by subsequently freeing and assigning one
          index = random.nextInt(size - assigned);
          cnt = 0;
          while ((assignment[cnt]) || (index > 0)) {
            if (!assignment[cnt]) index--;
            cnt++;
          }
          assignResource(cnt);
          index = random.nextInt(assigned);
          cnt = 0;
          while ((!assignment[cnt]) || (index > 0)) {
            if (assignment[cnt]) index--;
            cnt++;
          }
          freeResource(cnt);
        }
        else {  // mutate the whole assignment list
          for (index = 0; index < size; index++)  // first, free all assigned resources
            freeResource(index);
          while (assigned < min) {  // then, assign <tt>min==max</tt> new ones randomly
            index = random.nextInt(size - assigned);
            cnt = 0;
            while ((assignment[cnt]) || (index > 0)) {
              if (!assignment[cnt]) index--;
              cnt++;
            }
            assignResource(cnt);
          }
        }
      } // end of fixed case with min != size
      else { // this is a completely fixed list
        if (assigned != size)  // assign all elements
          for (int index = 0; index < size; index++)
            assignResource(index);
      }
    }
    else {  // just an ordinary list, no fixed stuff
      if (MUTATION_MODE == ONLY_ONE_OPERATION) {  // only perform one basic mutation
        int index, cnt;
        // first type of mutation: add one resource (upper limit: 2max)
        if ((assigned == 0) || ((assigned < size) && (assigned < 2*max) && (random.nextDouble() > 0.5))) {
          index = random.nextInt(size - assigned);
          cnt = 0;
          while ((assignment[cnt]) || (index > 0)) {
            if (!assignment[cnt]) index--;
            cnt++;
          }
          assignResource(cnt);
        }
        else {  // second type of mutation: free one resource (lower limit: 0)
          index = random.nextInt(assigned);
          cnt = 0;
          while ((!assignment[cnt]) || (index > 0)) {
            if (assignment[cnt]) index--;
            cnt++;
          }
          freeResource(cnt);
        }
      }
      else {  // perform multiple basic mutations
        free = getFreeResources();
        nonfree = getAssignmentList();
        Resource resource;
        int mutationCount, index;

        mutationCount = (max - min) / 2 + 1;  // perform that many basic mutation operations

        if (assigned > max) mutationCount = (assigned - max) + 1;  // or this many if we are outside the min-max-interval
        if (assigned < min) mutationCount = (min - assigned) + 1;

        while (mutationCount-- > 0) {
          // first type of mutation: add one resource (upper limit: 2max)
          if ((random.nextDouble() > 0.5) && (assigned < 2 * max) && (free.size() > 0)) {
            index = random.nextInt(free.size());
            resource = (Resource)free.get(index);
            assignResource(resource);
            nonfree.add(resource);
            free.remove(index);
          }
          // second type of mutation: remove one resource (lower limit: 0)
          if ((random.nextDouble() > 0.5) && (assigned > 0) && (nonfree.size() > 0)) {
            index = random.nextInt(nonfree.size());
            resource = (Resource)nonfree.get(index);
            freeResource(resource);
            free.add(resource);
            nonfree.remove(index);
          }
        }
      } // end of multiple operations case
    }
  } // end of void mutate()

  /** Returns the number of minimal assignments (lower boundary constraint). **/
  public final int getMin() {
    return min;
  }

  /** Returns the number of maximal assignments (upper boundary constraint). **/
  public final int getMax() {
    return max;
  }

  /** Lets the caller set extra information necessary for the timetabling process. **/
  public void setExtra(String value) {
    extra = value;
  }

  /** Returns the internal extra information. **/
  public String getExtra() {
    return extra;
  }
  
  public final String getName() {
    return name;
  }

  /** <tt>toString()</tt> just returns the name of the resource list. **/
  public String toString() {
    return name;
  }

  public int compareTo(java.lang.Object o) {
    return name.compareTo(o.toString());
  }
  
  public boolean equals(java.lang.Object o) {
    return (compareTo(o) == 0);
  }

  /** This is part of the <tt>EventtoXML()</tt> method which describes an <tt>Event</tt> instance in XML. **/
  public final String toXML() {
    int cnt = 0;
    String result = "  <ResourceList name=\"" + name + "\" min=\"" + min + "\" max=\"" + max + "\" assignment=\"";
    for (int i = 0; i < size; i++) if (assignment[i]) {
      Resource r = (Resource)takeFrom.get(i);
      result += GlobalMethods.plainToXML(r.name);
      if ((++cnt) < assigned)
        result += ",";
    }
    result += "\" takeFrom=\"";
    for (int i = 0; i < size; i++) {
      Resource r = (Resource)takeFrom.get(i);
      result += GlobalMethods.plainToXML(r.name);
      if (i < size - 1)
        result += ",";
    }
    return result + "\"/>\n";
  } // end of toXML()
  
} // end of class ResourceList

