/**
 * Implementation of class "Time".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;

/*****
 * The <tt>Time</tt> class manages the temporal aspects of a timetable.
 * An instance of the class <tt>Time</tt> can be of many different types
 * that describe either a point of time within a certain day or week or ...
 * Therefore, there are many different constructors, each with a slightly
 * different scope.
 *****/
public final class Time implements Comparable {
  private static final int TYPE_NOTYPE    = 0;
  private static final int TYPE_MINHOUR   = 1;
  private static final int TYPE_DAYTIME   = 2;
  private static final int TYPE_WEEKLY    = 3;
  private static final int TYPE_DATETIME  = 4;

  /** This is how time is modelled. **/
  public int minute, hour, day, week, month, year;

  /** A lazy constructor that sets everything to zero. **/
  public Time() {
    hour = minute = day = week = month = year = 0;
  }

  /**
   * This constructor takes a <tt>String</tt> of the form "Fr/17:50" or "Tu/9:30" or
   * "2001/09/11/11:47" or "1978/11/17/10:48"
   * and creates a <tt>Time</tt> from it. This is good for processing XML input.
   **/
  public Time(String aTime) {
    this();
    int index;
    aTime = aTime.toLowerCase();
    if ((aTime.charAt(0) >= 'a') && (aTime.charAt(0) <= 'z')) {  // week-day: "XX/hh:mm"
      index = aTime.indexOf('/');
      String day = aTime.substring(0, index);
      aTime = aTime.substring(index + 1);
           if (day.equals("mo")) this.day = 1;
      else if ((day.equals("tu")) || (day.equals("di"))) this.day = 2;
      else if ((day.equals("we")) || (day.equals("mi"))) this.day = 3;
      else if ((day.equals("th")) || (day.equals("do"))) this.day = 4;
      else if (day.equals("fr")) this.day = 5;
      else if (day.equals("sa")) this.day = 6;
      else if ((day.equals("su")) || (day.equals("so"))) this.day = 7;
      else {
        System.err.println("Error: \"" + day + "\" is unknown day name.");
        this.day = 0;
      }
      index = aTime.indexOf(':');
      hour = Integer.parseInt(aTime.substring(0, index));
      minute = Integer.parseInt(aTime.substring(index + 1));
    }
    else {  // ordinary date: YYYY/MM/DD/hh:mm
      List list = GlobalMethods.listFromString(aTime, '/');
      year = Integer.parseInt((String)list.get(0));
      month = Integer.parseInt((String)list.get(1));
      day = Integer.parseInt((String)list.get(2));
      if (list.size() > 3) {
        // finally, extract hour and minute
        aTime = (String)list.get(3);
        index = aTime.indexOf(':');
        if (index > 0) {
          hour = Integer.parseInt(aTime.substring(0, index));
          minute = Integer.parseInt(aTime.substring(index + 1));
        }
      }
    }  // end of "YYYY/MM/DD/hh:mm" case
  } // end of Time(String)

  /** This constructor can be used if the timetabling period is a single day or
    * part of it. **/
  public Time(int anHour, int aMinute) {
    this();
    hour = anHour;
    minute = aMinute;
  } // end of Time(int, int)

  /** This constructor can be used if the timetable to be created covers a sequence
    * of days within the same week or the same month. **/
  public Time(int aDay, int anHour, int aMinute) {
    this();
    day = aDay;
    hour = anHour;
    minute = aMinute;
  } // end of Time(int, int, int)

  /** This constructor can be used if the period of the timetable to be created
    * covers several weeks that shall be distinguished. **/
  public Time(int aWeek, int aDay, int anHour, int aMinute) {
    this();
    week = aWeek;
    day = aDay;
    hour = anHour;
    minute = aMinute;
  } // endof Time(int, int, int, int)

  /** This is the most general case where the timetabling period covers several years. **/
  public Time(int aYear, int aMonth, int aDay, int anHour, int aMinute) {
    this();
    year = aYear;
    month = aMonth;
    day = aDay;
    hour = anHour;
    minute = aMinute;
  } // end of Time(int, int, int, int, int)

  /** Performs a bijective transformation from <tt>Time</tt> to <tt>int</tt>. **/
  public final int toInt() {
    int result = minute + hour * 60 + day * 1440;
    result += week * 7 * 1440;
    result += month * 31 * 1440 + year * 12 * 31 * 1440;
    return result;
  }

  /** Compares to objects of type <tt>Time</tt>.
    * @return -1 if <tt>this<tt> is earlier than <tt>o</tt>, 0 if they are equal,
    *         +1 if <tt>o</tt> is earlier than <tt>this</tt> **/
  public final int compareTo(Object o) {
    int i1 = toInt();
    int i2 = ((Time)o).toInt();
    if (i1 < i2)
      return -1;
    if (i1 > i2)
      return +1;
    return 0;
  } // end of compareTo(Object)

  public final boolean equals(Object o) {
    try {
      if (((Time)o).toInt() == toInt())
        return true;
      return false;
    } catch (ClassCastException cce) {
      return false;
    }
  } // end of equals(Object)

  private String intToDay(int d) {
    switch (d) {
      case 1: return "Mo";
      case 2: return "Tu";
      case 3: return "We";
      case 4: return "Th";
      case 5: return "Fr";
      case 6: return "Sa";
      case 7: return "Su";
      default: return "Xx";
    }
  }

  /** Returns a string representation of the <tt>Time</tt> object. **/
  public String toString() {
    if (year < 0) {
      if (week < 0) {
        if (day < 0)
          return "" + hour + ":" + minute;
        else
          return "" + intToDay(day) + "/" + hour + ":" + minute;
      }
    }
    return "00:00";
  }

}// end of class Time

