/**
 * Implementation of class "TimeInterval".
 *
 * Author: Stefan Buettcher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 **/

package ttclasses;

import helperclasses.*;
import java.util.*;

/*****
 * A <tt>TimeInterval</tt> is the closed interval between a beginning and
 * an end (<tt>from</tt> and <tt>until</tt>), where beginning and end
 * are instances of <tt>Time</tt>. <br>
 * Note that there is no difference between a time interval and an "ordinary"
 * resource (such as a teacher or a room) from the problem's point of view.
 *****/
public final class TimeInterval extends Resource implements Comparable {
  /** The starting point of the interval. **/
  public Time from;
  /** The end point of the interval. **/
  public Time until;
  /** Determines if this interval is followed by a break. This is, for
   * instance, of interest if a school lesson is to be placed onto
   * two adjacent time slots. Then, the first slot may not be followed
   * by a break.
   **/
  public boolean breakFollows;
  /** This information is used to layout the timetable HTML output: Slots must be positioned. **/
  public int layoutX, layoutY, layoutZ;

  /** Just an ordinary constructor. **/
  public TimeInterval() {
    name = "(no name)";
    from = until = null;
    breakFollows = false;
    layoutX = layoutY = layoutZ = 0;
  }

  /** A constructor setting the name of the time interval. **/
  public TimeInterval(String name) {
    this.name = name;
    from = until = null;
    breakFollows = false;
    layoutX = layoutY = layoutZ = 0;
  }

  /** An ordinary constructor, setting the interval's beginning and end time. **/
  public TimeInterval(Time from, Time until) {
    this.from = from;
    this.until = until;
    this.breakFollows = false;
    name = "(no name)";
    layoutX = layoutY = layoutZ = 0;
  } // end of TimeInterval(Time, Time)

  /** A constructor that sets the interval's beginning and end time and defines whether this time interval
    * is followed by a break. **/
  public TimeInterval(Time from, Time until, boolean breakFollows) {
    this.from = from;
    this.until = until;
    this.breakFollows = breakFollows;
    name = "(no name)";
    layoutX = layoutY = layoutZ = 0;
  } // end of TimeInterval(Time, Time, boolean)

  public TimeInterval(Time from, Time until, String name) {
    this.from = from;
    this.until = until;
    this.name = name;
    layoutX = layoutY = layoutZ = 0;
  } // end of TimeInterval(Time, Time, String)

  public TimeInterval(Time from, Time until, boolean breakFollows, String name) {
    this.from = from;
    this.until = until;
    this.breakFollows = breakFollows;
    this.name = name;
    layoutX = layoutY = layoutZ = 0;
  } // end of TimeInterval(Time, Time, boolean, String)

  public void setVariable(String key, String value) {
    if (key.equals("from"))
      from = new Time(GlobalMethods.XMLtoPlain(value));
    else if (key.equals("until"))
      until = new Time(GlobalMethods.XMLtoPlain(value));
    else if (key.equals("layout")) {
      List l = GlobalMethods.listFromXMLString(value);
      layoutX = Integer.parseInt((String)l.get(0));
      layoutY = Integer.parseInt((String)l.get(1));
      layoutZ = Integer.parseInt((String)l.get(2));
    }
    else
      super.setVariable(key, value);
  } // end of setVariable(String, String)

  public final int compareTo(Object o) {
    return name.compareTo(o.toString());
  }

  /**
   * Checks if to time intervals are equal (i.e. their names are equal). This
   * states nothing about their beginnings or ends.
   **/
  public final boolean equals(Object o) {
    return (compareTo(o) == 0);
  }

  /**
   * The <tt>overlaps</tt> method checks if this interval is overlapped
   * by the one specified by <tt>ti</tt>.
   * @return true if the intervals overlap, false otherwise
   **/
  public final boolean overlaps(TimeInterval ti) {
    if (from.toInt() <= ti.from.toInt())
      return (until.toInt() > ti.from.toInt());
    else
      return (ti.until.toInt() > from.toInt());
  }
  
  /** Analogous to <tt>overlaps(TimeInterval)</tt>. **/
  public final boolean overlaps(TimeIntervals ti) {
    return ti.overlaps(this);
  }

  /** Returns the duration of the time interval in minutes. **/
  public final int getDuration() {
    return (until.toInt() - from.toInt());
  }

  /** Returns the distance (in minutes) between two time intervals (zero, if they are clashing). **/
  public final int distance(TimeInterval ti) {
    if (overlaps(ti))
      return -1;
    else
      if (from.toInt() < ti.from.toInt())
        return (ti.from.toInt() - until.toInt());
      else
        return (from.toInt() - ti.until.toInt());
  } // end of distance(TimeInterval)

  public void setLayout(int x, int y, int z) {
    layoutX = x;
    layoutY = y;
    layoutZ = z;
  } // end of setLayout(int, int, int)

  public String toString() {
    return name;
  }

  /** Returns the XML description of the TimeInterval element. **/
  public final String toXML(String type) {
    int j, k;
    String result = "<TimeSlot name=\"" + GlobalMethods.plainToXML(name) +
                    "\" type=\"" + GlobalMethods.plainToXML(type);
    result += "\" features=\"";
    for (j = 0; j < featureCount; j++) {
      result += GlobalMethods.plainToXML(features[j]);
      if (j < featureCount - 1)
        result += ",";
    }
    result += "\" ";
    result += "events=\"";
    for (j = 0; j < assignmentCount; j++) {
      result += GlobalMethods.plainToXML(assignments[j].event.name);
      if (j < assignmentCount - 1)
        result += ",";
    }
    result += "\" layout=\"" + layoutX + "," + layoutY + "," + layoutZ + "\"/>\n";
    return result;
  } // end of toXML(String)

} // end of class TimeInterval

